/** @file iop_sdevice.cpp Virtual device for interface definition  */

/*
   FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2008, Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

*/



#include "iop_sdevice.h"
#include "sys/time.h"
#include "sp_densityfnct.h"

// fix windows/mingw definition
#ifdef FAUDES_WINEXTRA
#define sleep(sec) Sleep((sec) * 1000)
#define usleep(usec) Sleep((usec) / 1000)
#endif

namespace faudes {


/*
 **********************************************
 **********************************************
 **********************************************

 implementation: AttributeSignalActuator

 **********************************************
 **********************************************
 **********************************************
 */

// faudes type std
FAUDES_TYPE_IMPLEMENTATION(AttributeSignalActuator,AttributeVoid)

//DoAssign(Src)
AttributeSignalActuator& AttributeSignalActuator::DoAssign(const AttributeSignalActuator& rSrcAttr) {
  FD_DHV("AttributeSignalActuator(" << this << "):DoAssign(): assignment from " <<  &rSrcAttr);
  mActions=rSrcAttr.mActions;
  return *this;
}

//DoWrite(rTw);
void AttributeSignalActuator::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) rLabel; (void) pContext;
  //dummy for token-io
  
  FD_DHV("AttributeSignalActuator::DoWrite()");
  rTw.WriteBegin("Actuator");
  rTw.WriteBegin("Actions");
  std::vector<Action>::const_iterator ait;

  for(ait=mActions.begin(); ait!=mActions.end(); ait++) {
    rTw << (Idx) ait->mBit;
    if(ait->mValue) rTw.WriteOption("Set");
    else rTw.WriteOption("Clr");
  }
  rTw.WriteEnd("Actions");
  rTw.WriteEnd("Actuator ");
}


//DoRead(rTr)
void AttributeSignalActuator::DoRead(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  (void) rLabel; (void) pContext;
  //dummy for token-io

  FD_DHV("AttributeSignalActuator::DoRead()");
  // clear
  mActions.clear();
  // test for actuator section
  Token token;
  rTr.Peek(token);
  if(token.Type()!=Token::Begin) return;
  if(token.StringValue()!="Actuator") return;
  // read actuator section
  rTr.ReadBegin("Actuator");
  bool err=false;
  rTr.ReadBegin("Actions");
  while(!rTr.Eos("Actions")) {
    // action
    Action action;
    // 1: bit address
    action.mBit=rTr.ReadInteger();
    // 2: value
    std::string value=rTr.ReadOption();
    if(value== "Set") {
      action.mValue=true;
    } else if(value == "Clr") {
      action.mValue=false;
    } else {
      err=true;
      break;
    }
    // 3. record
    mActions.push_back(action);
  }
  // done
  rTr.ReadEnd("Actions");
  rTr.ReadEnd("Actuator");
  // report error
  if(err) {
    std::stringstream errstr;
    errstr << "invalid actuator event property" << rTr.FileLine();
    throw Exception("AttributeSignalActuator::Read", errstr.str(), 52);  //52 oder 352
  }

}//DoRead


/*
 **********************************************
 **********************************************
 **********************************************

 implementation: AttributeSignalSensor

 **********************************************
 **********************************************
 **********************************************
 */


// faudes type std
FAUDES_TYPE_IMPLEMENTATION(AttributeSignalSensor,AttributeVoid)


//DoAssign(Src)
AttributeSignalSensor& AttributeSignalSensor::DoAssign(const AttributeSignalSensor& rSrcAttr) {
  FD_DHV("AttributeSignalSensor(" << this << "):DoAssign(): assignment from " <<  &rSrcAttr);
  mTriggers=rSrcAttr.mTriggers;
  return *this;
}


//DoWrite(rTw,rLabel,pContext)
void AttributeSignalSensor::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) rLabel; (void) pContext;
  //dummy for token-io

  FD_DHV("AttributeSignalSensor()::DoWrite()");
  rTw.WriteBegin("Sensor");
  rTw.WriteBegin("Triggers");
  std::vector<Trigger>::const_iterator ait;
  for(ait=mTriggers.begin(); ait!=mTriggers.end(); ait++) {
    rTw << ((Idx) ait->mBit);
    if(ait->mPos &&  (!ait->mNeg)) {
      rTw.WriteOption("PosEdge");
    }
    if((!ait->mPos) &&  ait->mNeg) {
      rTw.WriteOption("NegEdge");
    }
    if((ait->mPos) &&  ait->mNeg) {
      rTw.WriteOption("AnyEdge");
    }
  }
  rTw.WriteEnd("Triggers");
  rTw.WriteEnd("Sensor");
}//DoWrite(rTw,rLabel,pContext)


//DoRead(rTr,rLabel,pContext)
void AttributeSignalSensor::DoRead(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  (void) rLabel; (void) pContext;
  //dummy for token-io

  FD_DHV("AttributeSignalSensor()::DoRead(tr)");
  // clear
  mTriggers.clear();
  // test for sensor section
  Token token;
  rTr.Peek(token);
  if(token.Type()!=Token::Begin) return;
  if(token.StringValue()!="Sensor") return;
  // read sensor section
  rTr.ReadBegin("Sensor");
  bool err=false;
  rTr.ReadBegin("Triggers");
  while(!rTr.Eos("Triggers")) {
    // trigger
    Trigger trigger;
    // 1: bit address
    trigger.mBit=rTr.ReadInteger();
    // 2: value
    std::string value=rTr.ReadOption();
    if (value == "PosEdge") {
      trigger.mPos=true;
      trigger.mNeg=false;
    } else if (value == "NegEdge") {
      trigger.mPos=false;
      trigger.mNeg=true;
    } else if (value == "AnyEdge") {
      trigger.mPos=true;
      trigger.mNeg=true;
    } else {
      err=true;
      break;
    }
    // 3. record
    mTriggers.push_back(trigger);
  }
  // done
  rTr.ReadEnd("Triggers");
  rTr.ReadEnd("Sensor");
  // report error
  if(err) {
    std::stringstream errstr;
    errstr << "invalid sensor event property" << rTr.FileLine();
    throw Exception("AttributeSignalSensor::Read", errstr.str(), 52);
  }
}



/*
 **********************************************
 **********************************************
 **********************************************

 implementation: AttributeSignalEvent

 **********************************************
 **********************************************
 **********************************************
 */


// statics
AttributeSignalActuator AttributeSignalEvent::sActuatorPrototype;
AttributeSignalSensor AttributeSignalEvent::sSensorPrototype;

// std faudes type
FAUDES_TYPE_IMPLEMENTATION(AttributeSignalEvent,AttributeDeviceEvent)

// Default constructor, set my prototypes
AttributeSignalEvent::AttributeSignalEvent(void) : AttributeDeviceEvent() {
  FD_DHV("AttributeSignalEvent::AttributeSignalEvent(" << this << ")");
  pActuatorPrototype=&sActuatorPrototype;
  pSensorPrototype=&sSensorPrototype;
}

// Copy constructor 
AttributeSignalEvent::AttributeSignalEvent(const AttributeSignalEvent& rOtherAttr) : 
  AttributeDeviceEvent() 
{
  FD_DHV("AttributeSimplenetEvent(" << this << "): form other attr " <<  &rOtherAttr);
  pActuatorPrototype=&sActuatorPrototype;
  pSensorPrototype=&sSensorPrototype;
  DoAssign(rOtherAttr);
}


/*
 **********************************************
 **********************************************
 **********************************************

 implementation: sDevice

 **********************************************
 **********************************************
 **********************************************
 */


////////////////////////////////////////////////
// construction and destruction

//sDevice()
sDevice::sDevice(void) : vDevice() {
  // constructor

  FD_DHV("sDevice(" << mName << ")::sDevice()");
  // have event set with appropriate attributes
  mpConfiguration = new TaNameSet<AttributeSignalEvent>;
  pConfiguration = dynamic_cast< TaNameSet<AttributeSignalEvent>* >(mpConfiguration);
  // have appropriate default label for token io
  mDefaultLabel ="SignalDevice";
  // my signal data
  mMaxBitAddress=0;
  mpSensorEdges=0;
  mpRecentSensorEdges=0;
  mpActuatorLevels=0;
  // background thread:
  // install mutex
  pthread_mutex_init(&mMutex, NULL);
  // configuration defaults
  mName="SignalDevice";
  mCycleTime=1000; // 1ms
  mTimeScale=500;  // 500ms is one faudes time unit
  mSyncWrite=false;
}


//~sDevice()
sDevice::~sDevice(void) {
  // destructor

  FD_DHV("sDevice(" << mName << ")::~sDevice()");
  //cancel background thread: destroy mutex
  pthread_mutex_destroy(&mMutex);
  // invalidate data
  if(mpSensorEdges) delete mpSensorEdges;
  if(mpRecentSensorEdges) delete mpRecentSensorEdges;
  if(mpActuatorLevels) delete mpActuatorLevels;
}


//DoWritePreface(rTw,rLabel,pContext)
void sDevice::DoWritePreface(TokenWriter& rTw, const std::string& rLabel,  const Type* pContext) const {
  
  FD_DHV("sDevice("<<mName<<")::DoWritePreface()");
  //call base
  vDevice::DoWritePreface(rTw,rLabel,pContext);
  //write max bitaddress and cycle-time of background task
  rTw.WriteInteger(mMaxBitAddress);
  rTw.WriteInteger(mCycleTime);
  // write sync option
  if(mSyncWrite) rTw.WriteOption("SyncWrite");
}

//DoReadPreface(rTr,rLabel,pContext)
void sDevice::DoReadPreface(TokenReader& rTr,const std::string& rLabel, const Type* pContext){
  
  FD_DHV("sDevice("<<mName<<")::DoReadPreface()");
  // call base
  vDevice::DoReadPreface(rTr,rLabel,pContext);
  // read max bitaddress and cycle-time of background task
  mMaxBitAddress = rTr.ReadInteger();
  mCycleTime = rTr.ReadInteger();
  // figure sync option
  mSyncWrite=false;
  Token token;
  rTr.Peek(token);
  if(token.Type()==Token::Option)
  if(token.StringValue()=="SyncWrite") {
    rTr.Get(token);
    mSyncWrite=false;
  }
}





// lock - unlock shortcuts;
#define LOCK_E  {int rc = pthread_mutex_lock(&mMutex); \
  if(rc) {FD_ERR("sDevice("<<mName<<")::LOCK_E: lock mutex error\n"); exit(1); }}
#define UNLOCK_E  {int rc = pthread_mutex_unlock(&mMutex); \
  if(rc) {FD_ERR("sDevice("<<mName<<")::LOCK_E: unlock mutex error\n"); exit(1); }}


#define TLOCK_E  {int rc = pthread_mutex_lock(&sdevice->mMutex); \
  if(rc) {FD_ERR("sDevice::TLOCK_E: lock mutex error\n"); exit(1); }}

#define TUNLOCK_E  {int rc = pthread_mutex_unlock(&sdevice->mMutex); \
  if(rc) {FD_ERR("sDevice::TLOCK_E: unlock mutex error\n"); exit(1); }}





// Start(void)
void sDevice::Start(void) {
  //set device state and create thread

  //return if device was allready started
  if( (mState == Up)||(mState == StartUp) )  return;
  FD_DHV("sDevice("<<mName<<")::Start()");
  //call base incl virtual reset
  vDevice::Start();
  //create thread to handle input-signals
  if(!Sensors().Empty()){
    mCancelRequest = false;
    //setup thread attribute
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    //create thread
    int rc  = pthread_create(&mThreadSynchro, &attr, SDeviceSynchro, this);
    if (rc) {
      FD_ERR("sDevice("<<mName<<")::Start(): cannot creat thread\n");
      exit(1);
    }
    //destroy thread-attribute
    pthread_attr_destroy(&attr);
  }
}

// Stop()
void sDevice::Stop(void) {
  //set device state and kill thread

  //return if device allready got the command to shut down
  if(mState != Up && mState !=StartUp)  return;
  FD_DHV("sDevice("<<mName<<")::Stop()");
  //dont kill thread if no thread was created
  if(!Sensors().Empty()) {
    //cancle thread
    LOCK_E;
    mCancelRequest=true;
    UNLOCK_E;
    int pc = pthread_join(mThreadSynchro,NULL);
    if (pc) {
      FD_ERR("sDevice("<<mName<<")::Stop(): cannot join  thread??\n");
    }
  }
  // call base
  vDevice::Stop();
}//end Stop()


//Reset()
void sDevice::Reset(){
  //delete dynamic data
  FD_DHV("sDevice("<<mName<<")::Reset()");
  // call base (resets time)
  vDevice::Reset();
  //delete dynamic data
  ClrSensorSignals();
  ClrActuatorSignals();
}

// Clear()
void sDevice::Clear(void) {
  //delete static data

  FD_DHV("sDevice(" << mName << ")::Clear()");
  // call base; note: clear implies stop
  vDevice::Clear();
  // clear compiled data
  mSensorPosEdgeIndexMap.clear();
  mSensorNegEdgeIndexMap.clear();
  mMaxBitAddress=0;
}

//Compile(void)
void sDevice::Compile(void){
  //setup up internal data structure

  FD_DHV("sDevice(" << mName << ")::Compile()");
  // call base
  vDevice::Compile();
  // clear my internal data
  mSensorPosEdgeIndexMap.clear();
  mSensorNegEdgeIndexMap.clear();
  mMaxBitAddress=0;
  // setup my internal data
  for(EventSet::Iterator eit=pConfiguration->Begin(); eit!=pConfiguration->End(); eit++) {
    AttributeSignalEvent attr = pConfiguration->Attribute(*eit);
    if(attr.IsSensor()) {
      // fill in reverse maps
      std::vector<AttributeSignalSensor::Trigger>::const_iterator ait;
      const AttributeSignalSensor* sattr = attr.Sensorp();
      for(ait=sattr->mTriggers.begin(); ait!=sattr->mTriggers.end(); ait++) {
        if(ait->mPos) mSensorPosEdgeIndexMap[ait->mBit].Insert(*eit);
        if(ait->mNeg) mSensorNegEdgeIndexMap[ait->mBit].Insert(*eit);
        if(ait->mBit > mMaxBitAddress) mMaxBitAddress=ait->mBit;
      }
    }
    // check max address
    if(attr.IsActuator()) {
      std::vector<AttributeSignalActuator::Action>::const_iterator ait;
      const AttributeSignalActuator* aattr = attr.Actuatorp();
      for(ait=aattr->mActions.begin(); ait!=aattr->mActions.end(); ait++) {
        if(ait->mBit > mMaxBitAddress) mMaxBitAddress=ait->mBit;
      }
    }
  }
  // prep state data
  if(mpSensorEdges) delete mpSensorEdges;
  if(mpRecentSensorEdges) delete mpRecentSensorEdges;
  if(mpActuatorLevels) delete mpActuatorLevels;
  mpSensorEdges = new Edges[mMaxBitAddress+1];
  mpRecentSensorEdges = new Edges[mMaxBitAddress+1];
  mpActuatorLevels = new Levels[mMaxBitAddress+1];
}

//CycleTime(void)
int sDevice::CycleTime(void) const {
  //Report current cycle time
  
  FD_DHV("sDevice(" << mName << ")::CycleTime()");
  return mCycleTime;
}

//CycleTime(int)
void sDevice::CycleTime(int cycleTime) {
  //Set cycle time
  FD_DHV("sDevice(" << mName << ")::CycleTime(int)");
  if( mState==Up || mState==StartUp){
    std::stringstream errstr;
    errstr << "Changing cycle-time not possible while background thread is still running ";
    throw Exception("sDevice::CycleTime: ", errstr.str(), 100);
  }
  mCycleTime = cycleTime;
}


//ReadSignal(int)
bool sDevice::ReadSignal(int bit){
  FD_DH("sDevice(" << mName << ")::ReadSignal(int)");

  // make sure device is able to read signals
  if(Sensors().Empty()){
    std::stringstream errstr;
    errstr << "(" << mName << ") is not able to read signals";
    throw Exception("sDevice::ReadSignal():", errstr.str(), 552);  //TODO: number!
  }

  // run with hooks
  bool res=false;
  if(DoReadSignalsPre())
    res=DoReadSignal(bit);
  DoReadSignalsPost();

  return res;
}

//WriteSignal(int,int)
void sDevice::WriteSignal(int bit, bool value){
  // write one actor value, adressed by bit number (0 to 63);
  FD_DHV("sDevice(" << mName << ")::WriteSignal(int)");
  // make sure device supports actuators 
  if(Actuators().Empty()){
    std::stringstream errstr;
    errstr << "(" << mName << ") is not configured for actuators";
    throw Exception("sDevice::WriteSignal():", errstr.str(), 552);  
  }
  // lock global image, prevent background thread from reading/writing
  LOCK_E;
  // case a) asynchronous write
  if(!mSyncWrite) {
    if(DoWriteSignalsPre()) {
      DoWriteSignal(bit,value);
      DoWriteSignalsPost();
    }
  }
  // case b) synchronous write
  if(mSyncWrite) {
    // record next value
    Levels* level= &mpActuatorLevels[bit];
    level->next=value;
    if(level->edge) level->lost=true;
    level->edge=true;
  }
  // free global image
  UNLOCK_E;
}


//SDeviceSynchro(void*)
void* SDeviceSynchro(void* arg){
  //background-thread main-routine

  // cast this object
  sDevice* sdevice = static_cast<sDevice*>(arg);

  FD_DHV("sDevice(" << sdevice->mName << ")::Synchro(" << sdevice << "): with ct " << sdevice->mCycleTime);

	/////////////////////////////////////////////////////
	//	cyletime-measurement
	//	timeA: time consumed for a hole cycle
	//	timeB: time consumed by computation
	//	timeC: real time consumed by usleep
	////////////////////////////////////////////////////
#ifdef FAUDES_DEBUG_IOTIMING_X
	//arrays to store measured time in
	timeval* timeA = new timeval[MAX_SAMPLES];
	timeval* timeB = new timeval[MAX_SAMPLES];
	timeval* timeC = new timeval[MAX_SAMPLES];
        // define local itterator
	int itime=0;
	// take time
	gettimeofday(timeA+itime,0);
#endif

	bool running=true;

	while(running){

	  // lock global variables
	  TLOCK_E;

          // start up: initialze dynamic state
          if(sdevice->mState==sDevice::StartUp) {
            bool aready= sdevice->DoWriteSignalsPre();
   	    if(aready) {
              // actuator image: set to zero
              for(int bit=0; bit<=sdevice->mMaxBitAddress; bit++) {
	        FD_DHV("sDevice("<<sdevice->mName<<")::synchro: reset clear line");
		sdevice->DoWriteSignal(bit,false);
                sDevice::Levels* level= &sdevice->mpActuatorLevels[bit];
                level->current=false;
                level->next=false;
                level->edge=false;
                level->lost=false;
	      }
              sdevice->DoWriteSignalsPost();
            }
            bool sready= sdevice->DoReadSignalsPre();
   	    if(sready) {
              // edge detection: use actual signal levels
              for(int bit = 0;bit<=sdevice->mMaxBitAddress;bit++) {
                sDevice::Edges* edge= &sdevice->mpSensorEdges[bit];
	        FD_DHV("sDevice("<<mName<<")::synchro: reset edge detection");
                edge->current=sdevice->DoReadSignal(bit);
                edge->past = edge->current;
                edge->pos =    false;
                edge->neg =    false;
                edge->posrel = (!sdevice->mSensorPosEdgeIndexMap[bit].Empty());
                edge->negrel = (!sdevice->mSensorNegEdgeIndexMap[bit].Empty());
                edge->lost =    false;
              }
              sdevice->DoReadSignalsPost();
              for(int bit = 0;bit<=sdevice->mMaxBitAddress;bit++) {
                sDevice::Edges* redge= &sdevice->mpRecentSensorEdges[bit];
                redge->pos =    false;
                redge->neg =    false;
                redge->posrel = false;
                redge->negrel = false;
                redge->lost =   false;
              }
	    }
            // new state
            if(sready && aready) 
              sdevice->mState=sDevice::Up;   	    
	  }

          // normal operation, actuators
          if(sdevice->mState==sDevice::Up && sdevice->mSyncWrite) {
            bool ready=sdevice->DoWriteSignalsPre();
            if(!ready) 
              sdevice->mState=sDevice::StartUp;   
            if(ready) {
  	      // write values from image
	      for(int bit = 0; bit<=sdevice->mMaxBitAddress; bit++) {
	        // define pointer to levels
	        sDevice::Levels* level= &sdevice->mpActuatorLevels[bit];
	        // not wrtiien by libfaudes
	        if(!level->edge) continue;
	        // write and record
                sdevice->DoWriteSignal(bit,level->next);
	        level->current = level->next;
	        level->edge = false;
	      }
              sdevice->DoWriteSignalsPost();
	    }
	  }
 
          // normal operation, sensors
          if(sdevice->mState==sDevice::Up) {
            bool ready=sdevice->DoReadSignalsPre();
            if(!ready) 
              sdevice->mState=sDevice::StartUp;   
            if(ready) {
  	      // edge detection, accumulative
	      for(int bit = 0; bit<=sdevice->mMaxBitAddress; bit++) {
	        // define pointer to Sensoredges
	        sDevice::Edges* edge= &sdevice->mpSensorEdges[bit];
	        // pass edge-info about edge in last cycle
	        edge->past = edge->current;
	        // and get corresponding sensor-value
	        edge->current = sdevice->DoReadSignal(bit);
	        if(edge->posrel)
		if(edge->current && (!edge->past) ) {
		  /*
		    if(edge->pos && (!edge->lost)) {
		    FD_DH("sDevice::synchro: lost positive edge at bit address "  << bit);
		    edge->lost=true;
		    }
		  */
		  FD_DH("sDevice::synchro: sensed positive edge at bit address "  << bit);
		  edge->pos = true;
		  // queue events to buffer
		  pthread_mutex_lock(sdevice->pBufferMutex);
		  EventSet::Iterator eit=sdevice->mSensorPosEdgeIndexMap[bit].Begin();
		  EventSet::Iterator eit_end=sdevice->mSensorPosEdgeIndexMap[bit].End();
		  for(; eit!=eit_end; eit++) {
		    sdevice->pSensorBuffer->push_back(*eit);
		  }
		  pthread_mutex_unlock(sdevice->pBufferMutex);
		  //send signal to function "WaitSenosrs()"
		  FD_DH("sDevice::synchro: send signal " );
		  pthread_cond_broadcast(sdevice->pWaitCondition);
		}
	        if(edge->negrel)
		if( (!edge->current) && edge->past ) {
		  FD_DH("sDevice::synchro: sensed negative edge at bit address "  << bit);
		  edge->neg = true;
		  sdevice->mSensorReady = true;
		  // queue events to buffer
		  pthread_mutex_lock(sdevice->pBufferMutex);
		  EventSet::Iterator eit=sdevice->mSensorNegEdgeIndexMap[bit].Begin();
		  EventSet::Iterator eit_end=sdevice->mSensorNegEdgeIndexMap[bit].End();
		  for(; eit!=eit_end; eit++) {
		    sdevice->pSensorBuffer->push_back(*eit);
		  }
		  pthread_mutex_unlock(sdevice->pBufferMutex);
		  //send signal to function "WaitSenosrs"
		  FD_DH("sDevice::synchro: send signal " );
		  pthread_cond_broadcast(sdevice->pWaitCondition);
		}
	      } // loop bits
              sdevice->DoReadSignalsPost();
	    } // end-if device ok
	  } // end-if sensors
 
 
	  // check for cancel request
	  if(sdevice->mCancelRequest) running=false;
	  // unlock global variables
	  TUNLOCK_E;

	  //////////////////////////////////////
	  //cycletime measurement
	  //measure time to perform computation
#ifdef FAUDES_DEBUG_IOTIMING_X
	  if(itime < MAX_SAMPLES) gettimeofday(timeB+itime,0);
#endif

	  // let time pass; todo: use proper timer
	  usleep(sdevice->mCycleTime);

#ifdef FAUDES_DEBUG_IOTIMING_X
	  //////////////////////////////////////
	  // cycletime measurement

	  // measure time consumed by wait
	  if(itime < MAX_SAMPLES) 	gettimeofday(timeC+itime,0);
	  // measure time needed for a hole cyle
	  // increase iterator
	  itime++;
	  if(itime < MAX_SAMPLES) 	gettimeofday(timeA+itime,0);

#endif
	  // count cycles
	  sdevice->mCycleCount++;
	} // loop while running


#ifdef FAUDES_DEBUG_IOTIMING_X

	/////////////////////////////////////////////
	// cycletime measurement
	// analysis using SampleDensityFunction
      	FD_DHT("sDevice::synchro: performance analysis");

	SampledDensityFunction SamplesA;
	SamplesA.Clear();
        SamplesA.Dim(100);
        SamplesA.Name("time stamp AA: overall period");

	SampledDensityFunction SamplesB;
        SamplesB.Clear();
	SamplesB.Dim(100);
        SamplesB.Name("time stamp AB: process time");

	SampledDensityFunction SamplesC;
	SamplesC.Clear();
        SamplesC.Dim(100);
        SamplesC.Name("time stamp CB: waited time");

	SampledDensityFunction SamplesWS;
	SamplesWS.Clear();
	SamplesWS.Dim(100);
	SamplesWS.Name("time passed till the next call of WaitSensors");

	// timeval-structures needed for further computations
	timeval dAA,dBA,dCB,dER;

	// perform analysis
	// time for a hole cycle
	for(int ind = 0; ind < itime-2; ind++){
	  // compute time needed
	  // for a hole cycle
	  dAA = vDevice::DiffTime(timeA[ind+1],timeA[ind]);
	  // for coputing sensor-state
	  dBA = vDevice::DiffTime(timeB[ind],timeA[ind]);
	  // by usleep(mCyleTime)
	  dCB = vDevice::DiffTime(timeC[ind],timeB[ind]);
	  // time passed by until the next call of WaitSensors
	  dER = vDevice::DiffTime(sdevice->mptimeBegin[ind+1],sdevice->mptimeEnd[ind]);
	  // add computed samples to corresponding
	  // SampleDensityFunction classes
	  SamplesA.Sample(dAA.tv_sec*1000000 + dAA.tv_usec);
	  SamplesB.Sample(dBA.tv_sec*1000000 + dBA.tv_usec);
	  SamplesC.Sample(dCB.tv_sec*1000000 + dCB.tv_usec);
	  SamplesWS.Sample(1000000*dER.tv_sec + dER.tv_usec);
	}

	// perform statistic computation
        SamplesA.Compile();
        SamplesB.Compile();
	SamplesC.Compile();
	SamplesWS.Compile();

	// token output
	// SamplesA.Write();
	// SamplesB.Write();
	// SamplesC.Write();
	// SamplesWS.Write();

        // pretty output
        std::cout << SamplesA.Str() << std::endl;
        std::cout << SamplesB.Str() << std::endl;
        std::cout << SamplesC.Str() << std::endl;
	std::cout <<  SamplesWS.Str() << std::endl;
#endif
  FD_DH("sDevice(" << sdevice->mName << ")::synchro: terminate background thread");
  pthread_exit(NULL);
  // never happens, prevent gcc warning
  return NULL; 
}

// Write Actuator(Idx)
void sDevice::WriteActuator(Idx actuator) {

  FD_DHV("sDevice("<<mName<<")::WriteActuator");
  //return if device is running
  if(mState!=Up && mState!=StartUp) return;
  //make sure event is part of devices eventset
  if(!mActuators.Exists(actuator)) {
    std::stringstream errstr;
    errstr << "Unknown actuator event " << actuator;
    throw Exception("sDevice::WriteActuator", errstr.str(), 65);
  }
  // find properties
  const AttributeSignalActuator* aattr = pConfiguration->Attribute(actuator).Actuatorp();
  if(!aattr) {
    std::stringstream errstr;
    errstr << "Invalid actuator attribute " << actuator;
    throw Exception("sDevice::WriteActuator", errstr.str(), 65);
  }
  //debug-flag
  FD_DH("sDevice("<<mName<<")::WriteActuator: properties found");
  // execute command ...
  std::vector<AttributeSignalActuator::Action>::const_iterator eit;
  // ... case a) asynchronous write
  if(!mSyncWrite) {
    if(DoWriteSignalsPre()) {
      // loop actions
      for(eit=aattr->mActions.begin(); eit!=aattr->mActions.end(); eit++){
        FD_DH("sDevice("<<mName<<")::WriteActuator: mBit "<<eit->mBit<< " mValue "<<eit->mValue);
        DoWriteSignal(eit->mBit,eit->mValue);
      }
      DoWriteSignalsPost();
    }
  }
  // ... case b) synchronous write
  if(!mSyncWrite) {
    // lock global image
    LOCK_E;
    // loop actions and record
    for(eit=aattr->mActions.begin(); eit!=aattr->mActions.end(); eit++){
      FD_DH("sDevice("<<mName<<")::WriteActuator: mBit "<<eit->mBit<< " mValue "<<eit->mValue);
      Levels* level= &mpActuatorLevels[eit->mBit];
      level->next=eit->mValue;
      if(level->edge) level->lost=true;
      level->edge=true;
    }
    // free global image
    UNLOCK_E;
  }
  // done
  FD_DHV("sDevice("<<mName<<")::WriteActuator: done");
}

//ClrSensorSignals()
void sDevice::ClrSensorSignals(void) {
  FD_DHV("sDevice("<<mName<<")::ClrSensorSignals(): MaxBitAddress: " << mMaxBitAddress);
  //return if device is not running or device is not able to read sensors
  if(!(mState!=Up||Sensors().Empty())) return;
  // lock global variables
  FD_DHV("sDevice("<<mName<<")::ClrSensorSignals(): lock mutex ");
  LOCK_E;
  FD_DHV("sDevice("<<mName<<")::ClrSensorSignals(): edge detection ");
  // initialise edge detection: assume low signals
  for(int bit = 0;bit<=mMaxBitAddress;bit++) {
    Edges* edge= &mpSensorEdges[bit];
    FD_DHV("sDevice("<<mName<<")::ClrSensorSignals(): edge detetion ReadSignal bit "  << bit);
    edge->current=0;
    edge->past = 0;
    edge->pos =    false;
    edge->neg =    false;
    edge->posrel = (!mSensorPosEdgeIndexMap[bit].Empty());
    edge->negrel = (!mSensorNegEdgeIndexMap[bit].Empty());
    edge->lost =    false;
  }
  // initialise edge detection: actual signal levels
  if(DoReadSignalsPre()) {
    for(int bit = 0;bit<=mMaxBitAddress;bit++) {
      Edges* edge= &mpSensorEdges[bit];
      FD_DHV("sDevice("<<mName<<")::ClrSensorSignals(): edge detetion ReadSignal bit "  << bit);
      edge->current=DoReadSignal(bit);
      edge->past = edge->current;
      edge->pos =    false;
      edge->neg =    false;
      edge->posrel = (!mSensorPosEdgeIndexMap[bit].Empty());
      edge->negrel = (!mSensorNegEdgeIndexMap[bit].Empty());
      edge->lost =    false;
    }
    DoReadSignalsPost();
  }
  // initialize edge detection: recent edges
  FD_DHV("sDevice("<<mName<<")::ClrSensorSignals(): initialize recent edges ");
  for(int bit = 0;bit<=mMaxBitAddress;bit++) {
    Edges* redge= &mpRecentSensorEdges[bit];
    redge->pos =    false;
    redge->neg =    false;
    redge->posrel = false;
    redge->negrel = false;
    redge->lost =   false;
  }
  // reset cyclecount
  mCycleCount=0;
  mRecentCycleCount=0;
  mSensorReady=false;
  // unlock global variables
  UNLOCK_E;
  FD_DHV("sDevice("<<mName<<")::ClrSensorSignals(): done");
}

//ClrActuatorSignals()
void sDevice::ClrActuatorSignals(void) {
  //clear all actuator line levels
  FD_DHV("sDevice("<<mName<<")::ClrActuatorSignals");
  // return if device is not running or not abel to set actuator signals
  if(mState!=Up||Actuators().Empty()) return;
  // lock variables, prevent background from writing
  LOCK_E;
  // clear all (allways asynchronous)
  if(DoWriteSignalsPre()) {
    for(int bit=0; bit<=mMaxBitAddress; bit++)  DoWriteSignal(bit, false);
    DoWriteSignalsPost();
  }
  // clear image
  for(int bit=0; bit<=mMaxBitAddress; bit++) {
    Levels* level= &mpActuatorLevels[bit];
    level->current=false;
    level->next=false;
    level->edge=false;
    level->lost=false;
  }
  // done
  UNLOCK_E;
  FD_DHV("sDevice("<<mName<<")::ClrActuatorSignals: done");
}

}//end namespace faudes


