/** @file iop_vdevice.cpp Virtual device for interface definition  */

/*
   FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2008, Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

*/



#include "iop_vdevice.h"

#include "iop_simplenet.h"
#include "iop_comedi.h"
#include "iop_wago.h"
#include "iop_xdevice.h"
#include "sp_densityfnct.h"
#include "sys/time.h"


namespace faudes {



/*
 **********************************************
 **********************************************
 **********************************************

 implementation: AttributeDeviceEvent

 **********************************************
 **********************************************
 **********************************************
 */


// statics
AttributeVoid AttributeDeviceEvent::sFallbackAttribute=AttributeVoid();

// faudes type std
FAUDES_TYPE_IMPLEMENTATION(AttributeDeviceEvent,AttributeVoid)

// Default constructor (no attributes, aka undefined) */
AttributeDeviceEvent::AttributeDeviceEvent(void) :
  AttributeVoid(), 
  mpActuatorAttribute(0), 
  mpSensorAttribute(0),
  pActuatorPrototype(&sFallbackAttribute), 
  pSensorPrototype(&sFallbackAttribute)
{
  FD_DHV("AttributeDeviceEvent::AttributeDeviceEvent(" << this << ")");
}

// Copy constructor (never called)
AttributeDeviceEvent::AttributeDeviceEvent(const AttributeDeviceEvent& rOtherAttr) :
  AttributeVoid(), 
  mpActuatorAttribute(0), 
  mpSensorAttribute(0), 
  pActuatorPrototype(&sFallbackAttribute), 
  pSensorPrototype(&sFallbackAttribute)
{
  FD_DHV("AttributeDeviceEvent(" << this << "): form other attr " <<  &rOtherAttr);
  DoAssign(rOtherAttr);
}

// Destructor    
AttributeDeviceEvent::~AttributeDeviceEvent(void) {
  if(mpActuatorAttribute) delete mpActuatorAttribute;
  if(mpSensorAttribute)   delete mpSensorAttribute;
}

// Clear
void AttributeDeviceEvent::Clear(void) {
  if(mpActuatorAttribute) delete mpActuatorAttribute;
  if(mpSensorAttribute)   delete mpSensorAttribute;
  mpActuatorAttribute=0;
  mpSensorAttribute=0;
}

// Assignement
AttributeDeviceEvent& AttributeDeviceEvent::DoAssign(const AttributeDeviceEvent& rSrcAttr) {
  FD_DHV("AttributeDeviceEvent(" << this << "):DoAssign(): " 
    << typeid(*this).name() << " from " << typeid(rSrcAttr).name());
  Clear();
  if(rSrcAttr.IsSensor())   Sensor(*rSrcAttr.Sensorp());
  if(rSrcAttr.IsActuator()) Actuator(*rSrcAttr.Actuatorp());
  return *this;
}



// DoWrite(rTw,rLabel,pContext)
void AttributeDeviceEvent::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  FD_DHV("AttributeDeviceEvent(" << this << ")::DoWrite(tw): " << IsActuator() << " " << IsSensor()  );
  if(IsActuator()) mpActuatorAttribute->Write(rTw,rLabel,pContext);
  if(IsSensor())   mpSensorAttribute->Write(rTw,rLabel,pContext);
}

//DoRead(rTr,rLabel,pContext)
void AttributeDeviceEvent::DoRead(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  FD_DHV("AttributeDeviceEvent(" << this << ")::DoRead(tr)");
  // clear
  if(mpSensorAttribute) delete mpSensorAttribute;
  if(mpActuatorAttribute) delete mpActuatorAttribute;
  mpSensorAttribute=0;
  mpActuatorAttribute=0;
  // test for actuator/sensor section
  Token token;
  rTr.Peek(token);
  if(token.Type()!=Token::Begin) return;
  // read actuator/sensor
  if(token.StringValue()=="Actuator" || token.StringValue()=="Output") {
    DefaultActuator();
    mpActuatorAttribute->Read(rTr,rLabel,pContext);
    FD_DH("Found Actuator : " << mpActuatorAttribute->ToString());
  }
  // read actuator/sensor
  if(token.StringValue()=="Sensor" || token.StringValue()=="Input") {
    DefaultSensor();
    mpSensorAttribute->Read(rTr,rLabel,pContext);
    FD_DH("Found Sensor: " << mpSensorAttribute->ToString());
  }
}


/*
 **********************************************
 **********************************************
 **********************************************

 implementation: vDevice

 **********************************************
 **********************************************
 **********************************************
 */

// static members
std::set<vDevice*> vDevice::msAllDevices;

//vDevice()
vDevice::vDevice(void) {
  // constructor

  FD_DHV("vDevice(" << mName << ")::vDevice()");
  //set defaults
  mName = "IoDevice";				//device-name
  mpConfiguration = NULL;			//TODO: pointer to config-set
  mState = Down;				//initial state
  mDefaultLabel ="Device";			//label used with token-io
  mTimeScale = 1000;				//fauDES-time scaling-factor 1[ftu] =^ 1[sec]
  mMaxSyncGap = tpTime::Step;			//max time-delay used with simulator
  pWaitMutex=&mWaitMutex;			//use object - mutex
  pWaitCondition=&mWaitCondition; 		//use object - condition
  pBufferMutex=&mBufferMutex;			//use object - buffer-mutex
  pSensorBuffer=&mSensorBuffer; 		//use object - sensor-buffer

  //initial mutex- and condition variables
  pthread_mutex_init(&mWaitMutex,NULL);
  pthread_cond_init(&mWaitCondition,NULL);
  pthread_mutex_init(&mBufferMutex,NULL);

#ifdef FAUDES_DEBUG_IOPERF
  // initialize arrays to store time stamps
  mpPerformanceWaitEnter = new timeval[FAUDES_DEBUG_IOPERF_SAMPLES];
  mpPerformanceWaitExit = new timeval[FAUDES_DEBUG_IOPERF_SAMPLES];
  mPerformanceEndIterator = 0;
  mPerformanceBeginIterator = 0;
#endif

  // track devices
  msAllDevices.insert(this);

}

//~vDevice()
vDevice::~vDevice(void) {
  // destructor
  FD_DHV("vDevice(" << mName << ")::~vDevice()");
  //delete configuration
  if(mpConfiguration) delete mpConfiguration;
  //destroy mutex
  pthread_mutex_destroy(&mWaitMutex);
  pthread_cond_destroy(&mWaitCondition);
  pthread_mutex_destroy(&mBufferMutex);

#ifdef FAUDES_DEBUG_IOPERF
  // delete time stamps
  delete mpPerformanceWaitEnter;
  delete mpPerformanceWaitExit;
#endif

  // track devices
  msAllDevices.erase(this);
}

//FromTokenReader
vDevice* vDevice::FromTokenReader(TokenReader& rTr) {
  // construct from token reader
  FD_DHV("vDevice::FromTokenReader()");
  // peek first token // s&e tr->rTr
  Token token;
  rTr.Peek(token);
  if(!token.Type()==Token::Begin) {
    std::stringstream errstr;
    errstr << "Expected Begin Section in  \"" << rTr.FileName() << "\"";
    throw Exception("vDevice::FromFile", errstr.str(), 52);
  }
  // figure proper class and read
#ifdef FAUDES_IODEVICE_SIMPLENET
  // simple net device
  if(token.StringValue()=="SimplenetDevice") {
    FD_DHV("vDevice::FromFile(" << rTr.FileName() <<"): simplenet device");
    vDevice* ndevp=new nDevice();
    ndevp->Read(rTr);
    return ndevp;
  }
#endif
#ifdef FAUDES_IODEVICE_COMEDI
  // comedi device
  if(token.StringValue()=="ComediDevice") {
    FD_DHV("vDevice::FromFile(" << rTr.FileName() <<"): comedi device");
    cDevice* cdevp=new cDevice();
    cdevp->Read(rTr);
    return cdevp;
  }
#endif
#ifdef FAUDES_IODEVICE_WAGO
  // wago device
  if(token.StringValue()=="WagoDevice") {
    FD_DHV("vDevice::FromFile(" << rTr.FileName() <<"): wago device");
    wDevice* wdevp=new wDevice();
    wdevp->Read(rTr);
    return wdevp;
  }
#endif
  // device container aka xdevice
  if(token.StringValue()=="DeviceContainer") {
    FD_DHV("xDevice::FromFile(" << rTr.FileName() <<"): device container");
    xDevice* xdevp=new xDevice();
    xdevp->Read(rTr);
    return xdevp;
  }
  // unknown device type
  std::stringstream errstr;
  errstr << "Unknown type \"" << token.StringValue() << "\" at  \"" << rTr.FileLine()  << "\"";
  throw Exception("vDevice()::FromFile", errstr.str(), 52);
  return 0;
}


//FromFile(rFileName)
vDevice* vDevice::FromFile(const std::string& rFileName) {
  FD_DHV("vDevice::FromFile(" << rFileName <<")");
  // peek first token
  TokenReader tr(rFileName);
  return FromTokenReader(tr);
}


// static member function
void vDevice::StopAll(void) {
  // cosmetic bail out
  if(msAllDevices.size()==0) return;
  // report
  FD_DH("vDevice::StopAll()");
  // reset
  std::set<vDevice*>::const_iterator dit;
  for(dit=msAllDevices.begin(); dit != msAllDevices.end(); dit++) {
    if( dynamic_cast<xDevice*>(*dit) ) continue;
    FD_DH("vDevice::StopAll(): Reset " << *dit);
    (*dit)->Reset();
  }
  // stop
  for(dit=msAllDevices.begin(); dit != msAllDevices.end(); dit++) {
    if( dynamic_cast<xDevice*>(*dit) ) continue;
    FD_DH("vDevice::StopAll(): Stop " << *dit);
    (*dit)->Stop();
  }
}


// Start(void)
void vDevice::Start(void) {
  // only start if currently down
  if(mState!=Down) return;

  FD_DHV("vDevice(" << mName << ")::Start()");
  if(!mpConfiguration) {
    std::stringstream errstr;
    errstr << "cannot start device when not configured";
    throw Exception("vDevice()::Start()", errstr.str(), 550);
  }
  // optimistic default in base class
  mState=Up;
  // Start implies Reset
  Reset();
}


// Stop()
void vDevice::Stop(void) {
  // only stop if up or starting up
  if(mState!=Up && mState!=StartUp) return;
  FD_DHV("vDevice(" << mName << ")::Stop()");
  // Stop() implies Reset
  Reset();
  // optimistic default in base class
  mState=Down;
}


// Reset()
void vDevice::Reset(void){
  //delete dynamic data
  FD_DHV("vDevice(" << mName <<")::Reset()");
  //reset time and delete dynamic data
  CurrentTime(0);
  // reset sensor buffer
  pthread_mutex_lock(pBufferMutex);
  pSensorBuffer->clear();
  mResetRequest=false;
  pthread_mutex_unlock(pBufferMutex);
}


// Clear()
void vDevice::Clear(void) {
  //delete static data

  FD_DHV("vDevice("<< mName <<")::Clear()");
  //implies stop
  Stop();
  //clear exisiting data
  if(mpConfiguration) mpConfiguration->Clear();
  mActuators.Clear();
  mSensors.Clear();
}


// Name(rName)
void vDevice::Name(const std::string& rName) {
  FD_DHV("vDevice(" << mName << ")::Name(\"" << rName << "\")");
  mName = rName;
}

// Name()
const std::string& vDevice::Name(void) const {
  FD_DHV("vDevice(" << mName << ")::Name()");
  return mName;
}


//StatusString()
std::string vDevice::StatusString(void) {
  //status string, informal (should prepare mutex)

  FD_DHV("vDevice(" << mName << ")::Name()");
  //prepare result
  std::string res;
  //get curretn device state
  DeviceState currentState=Status();
  switch(currentState) {
  case Down: res="down"; break;
  case StartUp: res="starting up"; break;
  case Up: res="up and running"; break;
  case ShutDown: res="shutting down"; break;
  }

  return res;
}


//DoWrite(rTr,rLabel,pContext)
void vDevice::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) pContext; (void) rTw;
  //dummy for token output

  FD_DHV("vDevice("<<mName<<")::DoWrite()");
  //set default label if no label specified (configure in type constructor))
  std::string label = rLabel;
  if(label=="") label = mDefaultLabel;
  // write my section
  rTw.WriteBegin(label);
  //call DoWritePreface DoWriteConfiguration
  DoWritePreface(rTw,"",pContext);
  DoWriteConfiguration(rTw,"",pContext);
  // end of my section
  rTw.WriteEnd(label);
}

//DoWriteConfiguration(rTw,rLabel,pContext)
void vDevice::DoWriteConfiguration(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) pContext; (void) rTw;
  //dummy for token-input; writes configuration

  FD_DHV("vDevice("<<mName<<")::DoWriteConfiguration()");
  //Write label "EventConfiguration" if no label specified
  std::string label = rLabel;
  if(label == "") label = "EventConfiguration";
  //write event-configuration
  if(mpConfiguration) mpConfiguration->Write(rTw,rLabel);
}

//DoWritePreface(rTw,rLabel,pContext)
void vDevice::DoWritePreface(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) pContext; (void) rTw;
  //dummy for token-output; writes non-config data

  FD_DHV("vDevice("<<mName<<")::DoWritePreface()");
  //write device name and fauDES-time scaling-factor
  rTw.WriteString(mName);
  rTw.WriteInteger(mTimeScale);
}


//DoRead(rTr,rLabel,pContext)
void vDevice::DoRead(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  (void) pContext; (void) rLabel; (void) rTr;
  //dummy for token-input;

  FD_DHV("vDevice("<<mName<<")::DoRead()");
  Clear();
  //set default label if no label specified (configire in type constructor))
  std::string label = rLabel;
  if(label=="") label = mDefaultLabel;
  // read my section
  rTr.ReadBegin(label);
  //call DoReadConfiguration and DoReadPreface
  DoReadPreface(rTr,"",pContext);
  DoReadConfiguration(rTr,"",pContext);
  // end of my section
  rTr.ReadEnd(label);
  // organize internal data structures
  Compile();
}

//DoReadConfiguration(rTr,rLabel,pContext)
void vDevice::DoReadConfiguration(TokenReader& rTr,const std::string& rLabel, const Type* pContext){
  //dummy for token-input; reads configuration
  FD_DH("vDevice("<<mName<<")::DoReadConfiguration()");
  //set default label if no label specified
  std::string label=rLabel;
  if(label=="") label="EventConfiguration";
  //read event-configuration
  if(mpConfiguration) mpConfiguration->Read(rTr,label);
}

//DoReadPreface(rTr,rLabel,pContext)
void vDevice::DoReadPreface(TokenReader& rTr, const std::string& rLabel, const Type* pContext){
  (void) pContext; (void) rLabel; (void) rTr;
  //dummy for token-input; reads non-config data

  FD_DHV("vDevice("<<mName<<")::DoReadPreface()");
  // read name and fauDES-time scaling factor
  mName = rTr.ReadString();
  mTimeScale = rTr.ReadInteger();
}

// Configure(Idx,AttributeDeviceEvent&)
void vDevice::Configure(Idx event, const AttributeDeviceEvent& attr){
  //inserts an attributed event and compiles internal data structure 
 
  FD_DHV("vDevice(" << mName << ")::Configure(event)");
  Stop();
  // test type
  if(typeid(attr)!=typeid(mpConfiguration->Attribute())) {
    std::stringstream errstr;
    errstr << "Cannot cast attribute to device event attribute";
    throw Exception("vDevice::Configure", errstr.str(), 550);
  }
  // record
  mpConfiguration->Insert(event);
  mpConfiguration->Attribute(event,attr);
  // setup internal data structure
  Compile();
}

// Configure(EventSet)
void vDevice::Configure(const EventSet& rPhysicalEvents){
  FD_DHV("vDevice(" << mName << ")::Configure(alphabet)");
  Stop();
  // test type
  if(typeid(rPhysicalEvents.Attribute())!=typeid(mpConfiguration->Attribute())) {
    std::stringstream errstr;
    errstr << "Cannot cast attribute to device event attribute";
    throw Exception("vDevice::Configure", errstr.str(), 550);
  }
  // record events
  for(EventSet::Iterator eit=rPhysicalEvents.Begin(); eit!=rPhysicalEvents.End(); eit++) {
    const AttributeVoid attr = rPhysicalEvents.Attribute(*eit);
    mpConfiguration->Insert(*eit);
    mpConfiguration->Attribute(*eit,attr); // why does this do the job???
  }
  // setup internal data structure
  Compile();
}

//Compile()
void vDevice::Compile(void){
  FD_DHV("vDevice(" << mName << ")::Compile()");
  // Compile() implies Stop()
  Stop();
  // clear sensor- and actuator-set
  mSensors.Clear();
  mActuators.Clear();
  // loop over events and set actuators and sensors
  mSensors.Clear();
  mActuators.Clear();
  for(EventSet::Iterator eit=mpConfiguration->Begin(); eit!=mpConfiguration->End(); eit++) {
    AttributeVoid* attr = mpConfiguration->Attributep(*eit);
    AttributeDeviceEvent* dattr = dynamic_cast< AttributeDeviceEvent* >(attr);
    if(!dattr) {
      std::stringstream errstr;
      errstr << "Cannot cast attribute to device event attribute";
      throw Exception("vDevice()::Configure", errstr.str(), 550);
    }
    if(dattr->IsSensor()) mSensors.Insert(*eit);
    if(dattr->IsActuator()) mActuators.Insert(*eit);
  }
}

//UseCondition(pthread_mutex_t*)
void vDevice::UseCondition(pthread_mutex_t* wmutex, pthread_cond_t* wcond) {
  // tell device which condition variable to use
  FD_DHV("vDevice(" << mName << ")::UseCondition()");
  // set cond/mutex for wait sensors
  if(wmutex) pWaitMutex=wmutex;
  if(wcond) pWaitCondition=wcond;
}

//UseBuffer(mutex, buffer)
void vDevice::UseBuffer(pthread_mutex_t* bmutex, std::deque<Idx>* bbuffer) {
  // Tell the device which buffer to use
  FD_DHV("vDevice(" << mName << ")::UseBuffer()");
  //set sensor buffer and buffer-mutex and
  if(bmutex) pBufferMutex=bmutex;
  if(bbuffer) pSensorBuffer=bbuffer;
}

// Actuators()
const EventSet& vDevice::Actuators(void) const {
  return mActuators;
}

// Sensors()
const EventSet& vDevice::Sensors(void) const {
  return mSensors;
}

//DiffTime(timeval,timeval)
timeval vDevice::DiffTime(timeval end, timeval start){
  // computes differenz between two timval-structs
  FD_DHV("vDevice::DiffTime()");
  // declare a new timval construct to store the result in
  timeval res;
  // perform subtraction
  res.tv_sec = end.tv_sec-start.tv_sec;
  res.tv_usec = end.tv_usec-start.tv_usec;
  // and treat the carryover
  if(res.tv_usec <= 0){
    res.tv_sec--;
    res.tv_usec += 1000000;
  }
  return res;
}

//SumTime(timeval,timeval)
timeval vDevice::SumTime(timeval start, timeval duration){
  // computes sum of two timeval-structs
  FD_DHV("vDevice::SumTime()");
  // hypothesis: usec component is (to be kept) below 1000000
  timeval res;
  // perform addition
  res.tv_sec = start.tv_sec + duration.tv_sec;
  res.tv_usec = start.tv_usec + duration.tv_usec;
  // and treat the carryover
  if(res.tv_usec >= 1000000) {
    res.tv_sec++;
    res.tv_usec-=1000000;
  }
  return res;
}

//FtuToNsec(int)
timespec vDevice::FtuToTimeSpec(tpTime::Type duration){
  // convert fauDES-time to nsec(timespec)
  FD_DHV("vDevice("<<mName<<")::FtuToTimeSpec(" << duration << "): using scale " << mTimeScale);
  //prepare result
  timespec res;
  //convert duration to real time
  long int durationMsec = mTimeScale*duration;
  //get current system-time
  timeval now;
  gettimeofday(&now,0);
  // convert duration from mssec (long int) to usec (timeval)
  timeval delta;
  delta.tv_sec = durationMsec/1000;
  delta.tv_usec = ( durationMsec - ((durationMsec/1000)*1000)  ) *1000;
  //specify the absolut point of time
  timeval tmp_res = vDevice::SumTime(now,delta);
  // convert to timespec
  res.tv_sec = tmp_res.tv_sec;
  res.tv_nsec = tmp_res.tv_usec*1000;
  
  return res;
}

//FtuMsecToNsec(long int)
timespec vDevice::MsToTimeSpec(long int duration){
  //convert msec to nsec(timespec)
  FD_DHV("vDevice::MsToTimeSpec()");
  //prepare result
  timespec res;
  //get current system-time
  timeval now;
  gettimeofday(&now,0);
  // convert duration from msec (long int) to usec (timeval)
  timeval durationMsec;
  durationMsec.tv_sec = duration/1000;
  durationMsec.tv_usec = ( duration - ((duration/1000)*1000)  ) *1000;
  // has to wait before false is returned
  timeval tmp_res = vDevice::SumTime(now,durationMsec);
  // convertto timespec
  res.tv_sec = tmp_res.tv_sec;
  res.tv_nsec = tmp_res.tv_usec*1000;
  
  return res;
}

// FtuToMs(tpTime::Type)
long int vDevice::FtuToMs(tpTime::Type fauDES_time){
  //convert fauDES-time to realtime in ms

  FD_DHV("vDevice("<<mName<<")::FtuToMs() ");
  // avoid buffer-overflow while trying to convert an
  // "infinite" periode of time
  if(fauDES_time>= tpTime::Max/ (mTimeScale+1)) return tpTime::Max;
    // return real time in ms
  return mTimeScale*fauDES_time;
}

// MsToFtu(long int)
tpTime::Type vDevice::MsToFtu(long int real_time){
  //Convert real time in ms to fauDES-time

  FD_DHV("vDevice(" << mName << ")::MsToFtu()");
  return (long int)((1.0/mTimeScale)*real_time/1000*1000);
}

//CurrentTime(void)
tpTime::Type vDevice::CurrentTime(void) {
  //return current fauDES-time

  FD_DHV("vDevice("<<mName<<")::CurrentTime() ");
  // declarate a local timestructure
  struct timeval now;
  // and store the current system in it
  gettimeofday(&now,0);
  // compute the time elapsed since backgroundthread-startup
  timeval diffTime = vDevice::DiffTime(now,mTimeZero);
  // convert current physical time to current fauDES-time
  return (long int) ((1000.0/mTimeScale) * (diffTime.tv_sec + diffTime.tv_usec*1e-6));
}

//CurrentTime(now)
void vDevice::CurrentTime(tpTime::Type now){
  // set current fauDES-time

  FD_DHV("vDevice(" << mName <<")::CurrentTime("<<tpTime::Str(now) <<")");
  CurrentTimeMs(mTimeScale*now);
}

//CurrentTimeMs()
long int vDevice::CurrentTimeMs(void) {
  //return current real-time in ms

  FD_DHV("vDevice("<<mName<<")::CurrentTimeMs() ");
  // declare a local timestructure
  struct timeval now;
  // and store the current system in it
  gettimeofday(&now,0);
  // compute the time elapsed since backgroundthread-startup
  timeval diffTime = vDevice::DiffTime(now,mTimeZero);
  // convert current physical time to msecs
  return 1000*diffTime.tv_sec + diffTime.tv_usec/1000;
}


//CurrentTimeMs(long int)
void vDevice::CurrentTimeMs(long int nowms){
  // set current real-time in ms
  FD_DHV("vDevice("<<mName<<")::CurrentTimeMs(nowms) ");
  // set for now =0
  gettimeofday(&mTimeZero,0);
  // adjust to now
  timeval delta;
  delta.tv_sec = nowms/1000;
  delta.tv_usec = 1000*(nowms%1000);
  // carry
  if(delta.tv_usec >= 1000*1000){
    delta.tv_sec++;
    delta.tv_usec -= 1000000;
  }
  // do adjust
  mTimeZero=DiffTime(mTimeZero,delta);
  // report
  time_t systime;
  time(&systime);
  struct tm* timeinfo;
  timeinfo= localtime(&systime);
  char strbuf[80];
  strftime(strbuf,80,"%X",timeinfo);
}


// WaitSensors(tpTime::Type)
bool vDevice::WaitSensors(tpTime::Type duration) {
  //wait a specified time for sensor events to occur

  FD_DHV("vDevice("<<mName<<")::WaitSensors("<< duration << ")");

  // adjust arguments TODO
  if(duration> 3600*24) { duration= 3600*24;};

  // wait mutex gets released while waiting
  pthread_mutex_lock(pWaitMutex);

  // get preliminary result
  bool sr=SensorReady();
  // actually wait only if there are no sensor yet
  if(!sr && duration>0) {

#ifdef FAUDES_DEBUG_IOPERF
    // performance time stamp
    gettimeofday(mpPerformanceWaitEnter + mPerformanceEndIterator,0);
#endif

    //convert duration in nsec spec
    timespec condtime = vDevice::FtuToTimeSpec(duration);
    //wait for report from mThreadSynchro about occured sensor events
    FD_DH("vDevice("<<mName<<")::WaitSensors("<< duration << "): waiting for condition");
    pthread_cond_timedwait(pWaitCondition, pWaitMutex, &condtime);
    FD_DH("vDevice("<<mName<<")::WaitSensors("<< duration << "): release at "<< CurrentTime());
    // update result
    sr=SensorReady();

#ifdef FAUDES_DEBUG_IOPERF
    // performance time stamp
    gettimeofday(mpPerformanceWaitExit + mPerformanceEndIterator,0);
    mPerformanceEndIterator++;
    if(mPerformanceEndIterator >= FAUDES_DEBUG_IOPERF_SAMPLES) 
      mPerformanceEndIterator=0;
    if(mPerformanceEndIterator == mPerformanceBeginIterator)
      mPerformanceBeginIterator++;
    if(mPerformanceBeginIterator >= FAUDES_DEBUG_IOPERF_SAMPLES) 
      mPerformanceBeginIterator=0;
#endif

  }

  // release wait mutex
  pthread_mutex_unlock(pWaitMutex);


  return sr;
}


// WaitSensorsMs(long int)
bool vDevice::WaitSensorsMs(long int duration) {
  //wait a specified time for sensor events to occur

  FD_DHV("vDevice("<<mName<<")::WaitSensorsMs("<< duration << ")");

  // adjust arguments TODO
  if(duration> 30*3600*24) { duration= 30*3600*24;};

  // wait mutex gets released while waiting
  pthread_mutex_lock(pWaitMutex);

  // get preliminary result
  bool sr=SensorReady();
  // actually wait onnly if there are no sensor yet
  if(!sr && duration>0) {

#ifdef FAUDES_DEBUG_IOPERF
    // performance time stamp
    gettimeofday(mpPerformanceWaitEnter + mPerformanceEndIterator,0);
#endif

    //convert duration in nsec spec
    timespec condtime = vDevice::MsToTimeSpec( duration );
    //wait for report from mThreadSynchro about occured sensor events
    pthread_mutex_lock(pWaitMutex);
    pthread_cond_timedwait(pWaitCondition, pWaitMutex, &condtime);
    pthread_mutex_unlock(pWaitMutex);
    // update result
    sr=SensorReady();

#ifdef FAUDES_DEBUG_IOPERF
    // performance time stamp
    gettimeofday(mpPerformanceWaitExit + mPerformanceEndIterator,0);
    mPerformanceEndIterator++;
    if(mPerformanceEndIterator >= FAUDES_DEBUG_IOPERF_SAMPLES) 
      mPerformanceEndIterator=0;
    if(mPerformanceEndIterator == mPerformanceBeginIterator)
      mPerformanceBeginIterator++;
    if(mPerformanceBeginIterator >= FAUDES_DEBUG_IOPERF_SAMPLES) 
      mPerformanceBeginIterator=0;
#endif

  }

  // released wait mutex
  pthread_mutex_unlock(pWaitMutex);

  return sr;
}

// ResetRequest()
bool vDevice::ResetRequest(void) {
  bool res=false;
  pthread_mutex_lock(pBufferMutex);
  res=mResetRequest;
  mResetRequest=false;
  pthread_mutex_unlock(pBufferMutex);
  return res;
}

//ReadSensors()
Idx vDevice::ReadSensor(void) {
  //reads sensor-buffer and returns first occured events by index

  FD_DHV("vDevice("<<mName<<")::ReadSensor()");
  // prepare result, invalid
  Idx res=0;

  // take oldest entry
  pthread_mutex_lock(pBufferMutex);
  if(!pSensorBuffer->empty())  {
    res=pSensorBuffer->front();
    pSensorBuffer->pop_front();
  }
  pthread_mutex_unlock(pBufferMutex);
  FD_DHV("vDevice("<<mName<<")::ReadSensor(): found: " << res);
  return res;
}


//PeekSensors()
Idx vDevice::PeekSensor(void) {
  //reads sensor-buffer and returns first occured events by index
  // does not remove the event from the buffer

  FD_DHV("vDevice("<<mName<<")::PeekSensor()");
  // prepare result, invalid
  Idx res=0;

  // take oldest entry
  pthread_mutex_lock(pBufferMutex);
  if(!pSensorBuffer->empty())  {
    res=pSensorBuffer->front();
  }
  pthread_mutex_unlock(pBufferMutex);

  FD_DHV("vDevice("<<mName<<")::PeekSensor(): found: " << res);
  return res;
}


//SensorReady()
bool vDevice::SensorReady(void) {
  // returns true if a sensor-event occured

  FD_DHV("vDevice("<<mName<<")::ReadSensor()");
  //prepare result
  bool res;
  //read buffer
  pthread_mutex_lock(pBufferMutex);
  res= !mSensorBuffer.empty();
  pthread_mutex_unlock(pBufferMutex);

  return res;
}


// timing analysis
SampledDensityFunction vDevice::Performance(void) {
#ifdef FAUDES_DEBUG_IOPERF

  FD_DH("vDevice::Performance(" << Name() << "): " << mPerformanceBeginIterator << " " << mPerformanceEndIterator);

  // have samples
  SampledDensityFunction busyduration;
  busyduration.Dim(50);
  busyduration.Clear();
  busyduration.Name("performance " + Name() + " (busy duration in usecs)");

  // inspect time stamps
  for(int i = mPerformanceBeginIterator; ;i++){
    if(i>= FAUDES_DEBUG_IOPERF_SAMPLES) i=0;
    int ip = i+1;
    if(ip>= FAUDES_DEBUG_IOPERF_SAMPLES) ip =0;
    if(ip==mPerformanceEndIterator) break;
    timeval busy=vDevice::DiffTime(mpPerformanceWaitEnter[ip],mpPerformanceWaitExit[i]);
    busyduration.Sample(busy.tv_sec*1000000 + busy.tv_usec);
  }   

  // compile
  busyduration.Compile();

  // done
  return busyduration;

#else

  FD_DH("vDevice::Performance(" << Name() << "): not enabled");
  SampledDensityFunction busyduration;
  return busyduration;

#endif
}

// WritePerformance
void vDevice::WritePerformance(void) {

  SampledDensityFunction busyduration=Performance();

  // token output
  busyduration.Write();

  // pretty print
  std::cout << busyduration.Str() << std::endl;
}

// ClearPerformance
void vDevice::ResetPerformance(void) {

#ifdef FAUDES_DEBUG_IOPERF
  mPerformanceEndIterator = 0;
  mPerformanceBeginIterator = 0;
#endif

}

} // name space

