/** @file lbp_function.cpp luafaudes class to run scripts as rti functions */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2010 Thomas Moor

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

// my header
#include "lbp_function.h"

// all lua headers, incl luafaudes
#include "libluafaudes.h"



namespace faudes{

/*
********************************************************************
********************************************************************
********************************************************************

Implementation of class LuaFunctionDefinition

********************************************************************
********************************************************************
********************************************************************
*/

// faudes type
FAUDES_TYPE_IMPLEMENTATION(LuaFunctionDefinition,FunctionDefinition)

// construct
LuaFunctionDefinition::LuaFunctionDefinition(const std::string& name) : 
  FunctionDefinition(name) 
{
  Prototype(new LuaFunction(0));
  FD_DLB("LuaFunctionDefinition::LuaFunctionDefinition("<<this<<"): name   " << name);
  FD_DLB("LuaFunctionDefinition::LuaFunctionDefinition("<<this<<"): proto   " << mpFunction);
  FD_DLB("LuaFunctionDefinition::LuaFunctionDefinition("<<this<<"): has def " << mpFunction->Definition());
}

// copy construct
LuaFunctionDefinition::LuaFunctionDefinition(const LuaFunctionDefinition& rSrc) : 
  FunctionDefinition(rSrc.Name())
{
  FD_DLB("LuaFunctionDefinition::LuaFunctionDefinition(copy)");
  DoAssign(rSrc);
}

// std faudes type
LuaFunctionDefinition& LuaFunctionDefinition::DoAssign(const LuaFunctionDefinition& rSrc) {
  FD_DLB("LuaFunctionDefinition::DoAssign()");
  // assign base members
  FunctionDefinition::DoAssign(rSrc);
  // assign my members
  mLuaCode=rSrc.mLuaCode;
  mLuaFile=rSrc.mLuaFile;
  // special member
  pLuaFunction=dynamic_cast<LuaFunction*>(mpFunction);
  // report
  FD_DLB("LuaFunctionDefinition::DoAssign("<<this<<"): name   " << mName);
  FD_DLB("LuaFunctionDefinition::DoAssign("<<this<<"): proto   " << mpFunction);
  FD_DLB("LuaFunctionDefinition::DoAssign("<<this<<"): has def " << mpFunction->Definition())
  // done
  return *this;
}

// std faudes type
bool LuaFunctionDefinition::DoEqual(const LuaFunctionDefinition& rOther) const {
  // test base members
  if(!FunctionDefinition::DoEqual(rOther)) return false;
  // test my members
  if(mLuaCode!=rOther.mLuaCode) return false;
  if(mLuaFile!=rOther.mLuaFile) return false;
  return true;
}

// clear (all but prototype)
void LuaFunctionDefinition::Clear(){
  FD_DLB("LuaFunctionDefinition::Clear(): " << Name());
  // call base
  FunctionDefinition::Clear();
  // clear my data
  mLuaCode.clear();
}


// set prototype object
void LuaFunctionDefinition::Prototype(Function* pFunc){
  FunctionDefinition::Prototype(pFunc);
  // set typed version
  pLuaFunction=dynamic_cast<LuaFunction*>(pFunc);
}


// get lua code
const std::string& LuaFunctionDefinition::LuaCode(void) const {
  return mLuaCode;
}

// set lua code
void LuaFunctionDefinition::LuaCode(const std::string& rCode) {
  mLuaCode=rCode;
}


// test all variants
std::string LuaFunctionDefinition::SyntaxCheck(void) {
  // trivial case ... ok
  if(VariantsSize()==0) {
    FD_DLB("LuaFunctionDefinition::SyntaxCheck(): no variants defined, fine");
    return "";
  }    
  // allocate function
  faudes::LuaFunction* lfnct = pLuaFunction->New();
  // iterate variants
  std::string err = "";
  for(int i=0; i<lfnct->VariantsSize(); i++) {
    // select
    lfnct->Variant(i);
    FD_DLB("LuaFunctionDefinition::SyntaxCheck(): variant " << 
      lfnct->Variant()->Name());
    // allocate args (exception on unknown type, will lead to mem leak)
    try {
      lfnct->AllocateValues();
      // let luafunction have a go
      if(err=="") try {
        lfnct->SyntaxCheck();
      } catch(faudes::Exception ex)  {
        err = ex.What();
      }
      // free args
      lfnct->FreeValues();
    // alloc err
    } catch(faudes::Exception ex)  {
      err = "cannot allocate parameters for variant " + lfnct->Variant()->Name();
    }
  } // iterate variants
  delete lfnct;
  // done
  return err;
}


// token io
void LuaFunctionDefinition::DoRead(TokenReader& rTr,  const std::string& rLabel, const Type* pContext) {
  FD_DLB("LuaFunctionDefinition::DoRead()");
  // ignore
  (void) pContext;
  // label
  std::string label=rLabel;
  if(label=="") label="LuaFunctionDefinition";
  // base can handle this
  Documentation::DoRead(rTr,label,pContext);
  // report
  FD_DLB("LuaFunctionDefinition::DoRead(): done " << mPlugIn << "::" << mName);
}


// token io
void LuaFunctionDefinition::DoReadCore(TokenReader& rTr) {
  FD_DLB("LuaFunctionDefinition::DoReadCore()");
  // call base
  FunctionDefinition::DoReadCore(rTr);  
  // read my members
  Token token;
  rTr.Peek(token);
  // case a: embedded lua code
  if(token.Type()==Token::Begin)
  if(token.StringValue()=="LuaCode") {
    mLuaFile="";
    rTr.ReadBegin("LuaCode");
    mLuaCode=rTr.ReadString();
    rTr.ReadEnd("LuaCode");
  }
  // case b: lua file
  if(token.Type()==Token::Begin)
  if(token.StringValue()=="LuaFile") {
    rTr.ReadBegin("LuaFile");
    std::string mLuaFile=rTr.ReadString();
    // todo: read that file
    rTr.ReadEnd("LuaFile");
  }
}


// token io
void LuaFunctionDefinition::DoWrite(TokenWriter& rTw,  const std::string& rLabel, const Type* pContext) const {
   // label
  std::string label=rLabel;
  if(label=="") label="LuaFunctionDefinition";
  // base can handle
  Documentation::DoWrite(rTw,label,pContext); 
}

// token io
void LuaFunctionDefinition::DoWriteCore(TokenWriter& rTw) const {
  FD_DLB("LuaFunctionDefinition::DoWriteCore(): file " << rTw.FileName());
  // call base core
  FunctionDefinition::DoWriteCore(rTw);  
  // case a: embedded code
  if(mLuaFile=="") {
    rTw.WriteBegin("LuaCode");
    rTw.WriteVerbatim(mLuaCode);
    rTw.WriteEnd("LuaCode");
  }
  // case a: embedded code
  if(mLuaFile!="") {
    rTw.WriteBegin("LuaFile");
    rTw.WriteString(mLuaFile);
    rTw.WriteEnd("LuaFile");
    // todo: write that file
  }
}

/*
********************************************************************
********************************************************************
********************************************************************

Implementation of class LuaFunction

********************************************************************
********************************************************************
********************************************************************
*/

// Constructor 
LuaFunction::LuaFunction(const LuaFunctionDefinition* fdef) : 
  Function(fdef),
  pLuaFuncDef(fdef)
{
  FD_DLB("LuaFunction::LuaFunction(): fdef  " << pFuncDef);
}

// new on heap
LuaFunction* LuaFunction::New(void) const {
  return new LuaFunction(pLuaFuncDef);
}

// set function definition
void LuaFunction::Definition(const FunctionDefinition* fdef) {
  // cast and pass to base
  pLuaFuncDef = dynamic_cast<const LuaFunctionDefinition*>(fdef);
  Function::Definition(pLuaFuncDef);
}

// get function definition
const LuaFunctionDefinition* LuaFunction::Definition(void) const {
  return pLuaFuncDef;
}

// implement typecheck
bool LuaFunction::DoTypeCheck(int n) {
  FD_DLB("LuaFunction::DoTypeCheck("<< n << "): for " << Variant()->At(n).Type());
  const Type* proto = TypeRegistry::G()->Prototype(Variant()->At(n).Type()); 
  if(!proto) {
    FD_DLB("LuaFunction::DoTypeCheck("<< n << "): unknown type");
    return false;
  }
  if(!proto->Cast(ParamValue(n))) {
    FD_DLB("LuaFunction::DoTypeCheck("<< n << "): could not cast param value");
    return false;
  }
  return true;
}


// swigs version of user data (lua runtime version 4)
typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

typedef struct {
  swig_type_info   *type;
  int     own;  
  void        *ptr;
} swig_lua_userdata;





// implement execute: run stages
void LuaFunction::DoExecute(void) {
  FD_DLB("LuaFunction::DoExecute()");
  /*
  for(int i=0; i< Variant()->Size(); i++) {
    FD_DLB("LuaFunction::DoExecute(): value at #" << i);
    ParamValue(i)->Write();
  }
  */
  // have a state
  mpL = lua_open();    
  // run stages
  DoExecuteA();
  DoExecuteB();
  DoExecuteC();
  DoExecuteD();
  // clean stack
  lua_pop(mpL, 1); // table from stage A
  // done
  lua_close(mpL); 
  mpL=0;
}

// implement execute: syntaxcheck
void LuaFunction::SyntaxCheck(void) {
  FD_DLB("LuaFunction::SyntaxCheck()");
  // have a state
  mpL = lua_open();    
  // run stages
  DoExecuteA();
  DoExecuteB();
  DoExecuteD();
  // clean stack
  lua_pop(mpL, 1); // function from stage B
  lua_pop(mpL, 1); // table from stage A
  // done
  lua_close(mpL); 
  mpL=0;
}

// implement execute: get function on the stack
void LuaFunction::DoExecuteA(void) {
  FD_DLB("LuaFunction::DoExecuteA()");
  // prepare lua
  lua_gc(mpL, LUA_GCSTOP, 0);     // stop collector during initialization 
  luaL_openlibs(mpL);             // open libraries 
  luaopen_faudes_allplugins(mpL); // luafaudes: register
  //luafaudes_dotready();         // luafaudes: test dot
  lua_gc(mpL, LUA_GCRESTART, 0);  // start garbage collector
  // load my script
  const char* script = pLuaFuncDef->LuaCode().c_str();
  int script_len = pLuaFuncDef->LuaCode().size();
  int errload=luaL_loadbuffer(mpL, script, script_len, "luafaudes");
  if(errload!=0) {
    std::string lerr= std::string(lua_tostring(mpL, -1));
    int c1 = lerr.find_first_of(':');
    if(c1<0) c1=0;
    int c2 = lerr.find_first_of(':',c1+1);
    if(c2<0) c2=1;
    std::string line = lerr.substr(c1+1,c2-c1-1);  
    if(c2>1) {
      lerr="error in Lua script: line " + line + ": " + lerr.substr(c2+2);
    }
    lua_close(mpL); 
    mpL=0;    
    throw Exception("LuaFunction::DoExecuteA()", lerr, 49);
  }
  // run the script
  int errrun=lua_pcall(mpL, 0, 0, 0); 
  if(errrun!=0) {
    std::stringstream errstr;
    errstr << "failed to run script: ";
    errstr << std::string(lua_tostring(mpL, -1));
    lua_close(mpL); 
    mpL=0;
    throw Exception("LuaFunction::DoExecuteA()", errstr.str(), 49);
  }
  // get the swig generated table of faudes functions
  lua_getglobal(mpL,"faudes");
  mFtable=lua_gettop(mpL);
  if(!lua_istable(mpL,-1)) {
    std::stringstream errstr;
    errstr << "failed to load faudes table";
    lua_close(mpL);
    mpL=0; 
    throw Exception("LuaFunction::DoExecuteA()", errstr.str(), 49);
  }
  // mangle my function name
  std::string fname = Variant()->Name();
  unsigned int src=0;
  unsigned int dst=0;
  while(src<fname.length()) {
    if(isalnum(fname.at(src))) {fname[dst++]=fname[src++]; continue;}
    fname[dst++]='_';
    for(;src<fname.length();src++) 
      if(isalnum(fname.at(src))) break;
  }
  fname.resize(dst);
  // find my function
  lua_getglobal(mpL,fname.c_str()); 
  if(!lua_isfunction(mpL,-1)) {
    std::stringstream errstr;
    errstr << "missing function \"" << fname << "\"";
    lua_close(mpL);
    mpL=0; 
    throw Exception("LuaFunction::DoExecuteA()", errstr.str(), 49);
  }
}

// implement execute: prepare parameters
void LuaFunction::DoExecuteB(void) {
  FD_DLB("LuaFunction::DoExecuteB()");
  // construct my parameters in lua
  for(int i=0; i< Variant()->Size(); i++) {
    const std::string& ftype= Variant()->At(i).Type();
    // special case: int as lua number
    if(ftype=="Integer") {
      lua_pushnumber(mpL,((Integer*)ParamValue(i))->CValue());
      FD_DLB("LuaFunction::DoExecuteB(): created type " << ftype);
      continue;
    }
    // special case: bool as lua bool
    if(ftype=="Boolean") {
      lua_pushboolean(mpL,((Boolean*)ParamValue(i))->CValue());
      FD_DLB("LuaFunction::DoExecuteB(): created type " << ftype);
      continue;
    }
    // special case: str as lua string
    if(ftype=="String") {
      lua_pushstring(mpL,((String*)ParamValue(i))->CValue().c_str());
      FD_DLB("LuaFunction::DoExecuteB(): created type " << ftype);
      continue;
    }
    // std case: faudes type: construct 1
    lua_pushstring(mpL,ftype.c_str());
    lua_gettable(mpL,mFtable);
    if(!lua_isfunction(mpL,-1)) {
      std::stringstream errstr;
      errstr << "failed load constructor for \"" << ftype << "\"";
      lua_close(mpL);
      mpL=0; 
      throw Exception("LuaFunction::DoExecuteB()", errstr.str(), 49);
    }
    // std case: faudes type: construct 2
    if(lua_pcall(mpL, 0, 1, 0) != 0) { 
      std::stringstream errstr;
      errstr << "failed to construct for \"" << ftype << "\" (1)";
      lua_close(mpL);
      mpL=0; 
      throw Exception("LuaFunction::DoExecuteB()", errstr.str(), 49);
    }
    swig_lua_userdata* susr = (swig_lua_userdata*) lua_touserdata (mpL, -1);
    // std case: test pointer
    if(!susr) {
      std::stringstream errstr;
      errstr << "failed to construct for\"" << ftype << "\" (2)";
      lua_close(mpL);
      mpL=0; 
      throw Exception("LuaFunction::DoExecuteB()", errstr.str(), 49);
    }
    FD_DLB("LuaFunction::DoExecuteB(): created type " << susr->type->name);
    // std case: reown to save rom garbadge collector
    if(susr->own) free(susr->ptr);
    susr->ptr = ParamValue(i);
    susr->own=0;
  }
}

// implement execute: execute the function
void LuaFunction::DoExecuteC(void) {
  FD_DLB("LuaFunction::DoExecuteC()");
  // run my function
  if(lua_pcall(mpL, Variant()->Size(), Variant()->Size(), 0) != 0) { 
    std::string lerr= std::string(lua_tostring(mpL, -1));
    // user request ?
    if(lerr.find("break on application request")!=std::string::npos)
      lerr="break on application request";
    // have line number ?
    int c1 = lerr.find_first_of(':');
    if(c1<0) c1=0;
    int c2 = lerr.find_first_of(':',c1+1);
    if(c2<0) c2=1;
    std::string line = lerr.substr(c1+1,c2-c1-1);  
    if(c2>1) {
      lerr="error in Lua script: line " + line + ": " + lerr.substr(c2+2);
    }
    lua_close(mpL); 
    mpL=0;
    throw Exception("LuaFunction::DoExecuteC()", lerr, 49);
  }
}

// implement execute: retrieve results
void LuaFunction::DoExecuteD(void) {
  FD_DLB("LuaFunction::DoExecuteD()");
  // retrieve results from stack
  for(int i=Variant()->Size()-1; i>=0; i--) {
    const std::string& ftype= Variant()->At(i).Type();
    // special case: int as lua number
    if(ftype=="Integer" && lua_isnumber(mpL,-1)) {
      FD_DLB("LuaFunction::DoExecuteD(): retrieve type " << ftype);
      ((Integer*)ParamValue(i))->CValue(lua_tonumber(mpL,-1));
      lua_pop(mpL, 1); 
      continue;
    }
    // special case: bool as lua bool
    if(ftype=="Boolean" && lua_isboolean(mpL,-1)) {
      ((Boolean*)ParamValue(i))->CValue(lua_toboolean(mpL,-1));
      FD_DLB("LuaFunction::DoExecuteD(): created type " << ftype);
      lua_pop(mpL, 1); 
      continue;
    }
    // special case: str as lua string
    if(ftype=="String" && lua_isstring(mpL,-1)) {
      ((String*)ParamValue(i))->CValue(lua_tostring(mpL,-1));
      FD_DLB("LuaFunction::DoExecuteD(): created type " << ftype);
      lua_pop(mpL, 1); 
      continue;
    }
    // std case: faudes types are dealt with by reference anyway
    if(lua_isuserdata(mpL,-1)) {
      FD_DLB("LuaFunction::DoExecuteD(): retrieve type " << ftype);
      lua_pop(mpL, 1); 
      continue;
    }
    // report error
    std::stringstream errstr;
    errstr << "invalid return values in \"" << Variant()->Name() << "\"";
    lua_close(mpL); 
    mpL=0;
    throw Exception("LuaFunction::DoExecuteD()", errstr.str(), 49);
  }
} 


} // namespace

