/** @file lbp_function.h luafaudes class to run scripts as rti functions */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2010 Thomas Moor

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef FAUDES_LBPFUNCTION_H
#define FAUDES_LBPFUNCTION_H

#include "corefaudes.h"

// forward
struct lua_State;

namespace faudes{

class LuaFunction;

/**
 * A LuaFunctionDefinition is derived from FaunctionDefinition to
 * defines a faudes-function to be a luafaudes script. Thus, it has
 * the lua code as an additional member variable and uses a LuaFunction as a
 * prototype object. In particular, LuaFunctionDefinitions are allways
 * valid in the sense that hey have a prototype. The LuaFunction object 
 * implements DoTypeCheck and DoExecute to run the specifed lua code. 
 *
 * The token io format is demonstrated by the
 * following example:
 *
 * @code
 * <LuaFunctionDefinition>
 *
 * "LuaInteregrSum::LuaExtension"
 *
 * <TextDoc> "Returns the sum of two integer arguments." </TextDoc>
 * <HtmlDoc> "integersum.html" </HtmlDoc>
 * <Keywords> "luafunctions" "example" </Keywords>
 *
 * <VariantSignatures>
 * <Signature>
 * "Res=A+B"
 * "ArgA"  "Integer"   +In+
 * "ArgB"  "Integer"   +In+
 * "Res"   "Integer"   +Out+
 * </Signature>
 * </VariantSignatures>
 * 
 * <LuaCode> 
 * __VERBATIM__
 *
 * -- Define my function
 * function Res_A_B(a,b,res)
 *   print(string.format('Res_A_B(...): a=%d b=%d',a,b))
 *   res=a+b
 *   return a,b,res
 * end
 *
 * __VERBATIM__
 * </LuaCode>
 *
 * </LuaFunctionDefinition>
 * @endcode
 *
 * Alternatively to the embedded code, the luacode can be read from an external file, specified
 * by a section <tt>LuaFile</tt>. (not implemented)
 *
 * Restrictions and conventions:
 * - Typechecking is done via the Cast() function of the faudes type interface. As a 
 *   consequence, you may only use types that are registered with run-time-interface.
 * - On execution, the LuaFunction instance has to locate the respective function 
 *   in the supplied lua code. In order to allow for multiple variants, the convention is
 *   to have one lua function each with the name of the corresponding variant. Since
 *   variant names may contain funny characters, name matching is performed after
 *   replacing any sequence of non-alpha-numericals by "_". E.g. the variant "res=a+b" matches the
 *   lua function "res_a_b".
 * - Since luafaudes has no concept of const references, it is the responsability of the
 *   script programer to respect parameter types +In+, +Out+ and +InOut+.
 * - Integers, booleans ans strings are passed to Lua by value, any other parameters
 *   are passed by reference. To have a uniform interface, the convention is that the 
 *   lua function must return all specified parameters.
 *
 *
 *
 * @ingroup LuabindingsPlugin
 */

class LuaFunctionDefinition : public FunctionDefinition {

  // faudes type
  FAUDES_TYPE_DECLARATION(LuaFunctionDefinition,FunctionDefinition)

public:

  /** 
   * Constructor 
   *
   * In contrast to the std FunctionDefinition, the default constructor 
   * sets up a valid lua function definition with a newly created LuaFunction 
   * as prototype.
   * Of course, you will need to set the signatures and the lua code
   * to obtain an operational function.
   */
  LuaFunctionDefinition(const std::string& name="");

  /** 
   * Copy constructor 
   */
  LuaFunctionDefinition(const LuaFunctionDefinition& rSrc);

  /**
   * Destructor
   */
  virtual ~LuaFunctionDefinition(void){};


  /**
   * Clear documentation-data, signature and script (keep prototype)
   */
  void Clear(void);

  /**
   * Get Lua code
   * 
   * @return
   *   Lua code as std string
   */
  const std::string& LuaCode(void) const;

  /**
   * Set Lua code
   * 
   * @param rCode
   *   Lua code as std string
   */
  void LuaCode(const std::string& rCode);


  /**
   * Syntax check lua code.
   *
   * This routine instantiates a LuaFunction from this function definition
   * and does all it needs to run the script, except to invoke the any of the
   * variant functions. The reasoning is, that the script may hang and, thus, 
   * never return. Errors are indicated by an exception.
   * 
   * @return
   *   Error message as string, or empty string on success
   */
  std::string SyntaxCheck(void);

protected:

  /**
   * Std faudes type interface: assignment.
   *
   * @param rSrc 
   *    Source to copy from
   * @return Reference to this object.
   */
  virtual LuaFunctionDefinition& DoAssign(const LuaFunctionDefinition& rSrc);

  /**
   * Std faudes type interface: test equality
   * 
   * @param rOther 
   *    Other object to compare with.
   * @return 
   *   True on match.
   */
  virtual bool DoEqual(const LuaFunctionDefinition& rOther) const;

  /**
   * Read configuration data of this object from TokenReader.
   * Actual reading is done by DoReadCore.
   *
   * The section defaults to "LuaFunctionDefinition", context ignored.
   *
   * @param rTr
   *   TokenReader to read from
   * @param rLabel
   *   Section to read
   * @param pContext
   *   Read context to provide contextual information (ignored)
   *
   * @exception Exception
   *   - Token mismatch (id 50, 51, 52)
   *   - IO error (id 1)
   */
  virtual void DoRead(TokenReader& rTr,  const std::string& rLabel = "", const Type* pContext=0);
 
  /**
   * Read configuration data of this object from TokenReader.
   *
   * This method reads members only, it does not read the section.
   *
   * @param rTr
   *   TokenReader to read from
   *
   * @exception Exception
   *   - Token mismatch (id 50, 51, 52)
   *   - IO error (id 1)
   */
  virtual void DoReadCore(TokenReader& rTr);
 
  /**
   * Write configuration data of this object to TokenWriter.
   *
   * The section defaults to "LuaFunctionDefinition", context ignored.
   *
   * @param rTw
   *   Reference to TokenWriter
   * @param rLabel
   *   Label of section to write
   * @param pContext
   *   Write context to provide contextual information
   *
   * @exception Exception 
   *   - IO errors (id 2)
   */
  virtual void DoWrite(TokenWriter& rTw, const std::string& rLabel="",const Type* pContext=0) const;

  /**
   * Write configuration data of this object to TokenWriter.
   *
   * This method writes plain member data, the section lables are not
   * written.
   *
   * @param rTw
   *   Reference to TokenWriter
   *
   * @exception Exception 
   *   - IO errors (id 2)
   */
  virtual void DoWriteCore(TokenWriter& rTw) const;


  /**
   * Assign prototype object
   *
   * @param pFunc
   * 	Function instance
   *
   */
  virtual void Prototype(Function* pFunc);

  /** Typed prototype instance */
  LuaFunction* pLuaFunction;

  /** Lua code */
  std::string mLuaCode;

  /** Lua file */
  std::string mLuaFile;

}; 


/**
 * A LuaFunction is a faudes-function that executes a luafaudes script. 
 *
 * LuaFunction is derived from Function and implements the DoTypeCheck and DoExecute
 * interface to run the lua code as supplied by the corresponding function defintion.
 * Thus, it is considered an error to set the function definition to an object that
 * does not cast to a LuaFunctionDefinition.
 *
 * @ingroup LuabindingsPlugin
 */  

class LuaFunction : public Function {

 public:
  /** 
   * Constructor 
   * For the function to be operational, a valid reference to the corresponding 
   * FunctionDefinition is required. The only exception is the prototype function
   * object used in the FunctionDefinition itself. 
   */
  LuaFunction(const LuaFunctionDefinition* fdef);

  /** Destructor */
  ~LuaFunction(void){};

  /**
   * Construct on heap.
   * Create a new instance of this function class and return pointer.
   * The new instance will use the same function definition as this instance.
   *
   * @return
   * 	Pointer to faudes::Function instance.
   *
   */
  virtual LuaFunction* New() const;


  /**
   * Set function definition.
   * Normally, functions are provided with a function definition on construction.
   * The only exception are prototype objects used in function definitions themselfs 
   * and in the function registry.
   *
   * @param fdef
   * 	Function definition to set.
   *
   */
  void Definition(const FunctionDefinition* fdef);


  /**
   * Get function definition.
   *
   * @return 
   * 	Function definition used by this function.
   *
   */
  const LuaFunctionDefinition* Definition(void) const;


  /**
   * Syntax check lua code.
   *
   * This routine does all it needs to run the script,
   * except to invoke the specified function. The reasoning is, that
   * the script may hang and, thus, never return. A consequence
   * is, that you must set a variant and you must supply parameter 
   * values befor checking. You may use AllocateValues() and FreeValues() 
   * for this purpose. Errors are indicated by an exception.
   *
   * Note that the LuaFunctionDefinition provides a convenience wrapper
   * that runs the check on all variants and cares about value allocation.
   * 
   * @exception Exception
   *   - Error in Lua script (id 49)
   */
  void SyntaxCheck(void);


 protected:

  /**
   * Method to test the type of an assigned parameter with the
   * specified faudes::Signature (i.e. their TypeDefinition label).
   *
   * Note: this method is called by Function::Execute() before actual function 
   * execution via DoExecute(). It may be used to perform a dynamic cast in 
   * preparation of DoExecute(). The latter is only called, if all types match.
   * 
   * @param n
   *   Position of parameter to check
   * @return
   *   True if type matches signature.
   *
   * @exception Exception
   * 	- Signature undefined (id 48)
   * 	- Parameter number out of range (id 48)
   */
  virtual bool DoTypeCheck(int n);

  /**
   * Executes code as supplied by FunctionDefinition
   *
   * @exception Exception
   * 	- Exception during lua setup (id 49)
   * 	- Any exception during execution of script
   */
  virtual void DoExecute();

  /**
   * Execute stages
   *
   * @exception Exception
   * 	- Exception during lua setup (id 49)
   */
  virtual void DoExecuteA();

  /**
   * Execute stages
   *
   * @exception Exception
   * 	- Exception during lua setup (id 49)
   */
  virtual void DoExecuteB();

  /**
   * Execute stages
   *
   * @exception Exception
   * 	- Exception during lua setup (id 49)
   * 	- Any exception during execution of script
   */
  virtual void DoExecuteC();

  /**
   * Execute stages
   *
   * @exception Exception
   * 	- Exception during lua setup (id 49)
   */
  virtual void DoExecuteD();

  /** Typed reference to definition */
  const LuaFunctionDefinition* pLuaFuncDef;

  /** State of Lua interpreter */
  lua_State* mpL;
  int mFtable;
}; 


} // namespace
#endif // include
