/** @file syn_supcon.cpp Supremal controllable sublanguage */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */
   

#include "syn_supcon.h"


namespace faudes {




/*
***************************************************************************************
***************************************************************************************
 Implementation
***************************************************************************************
***************************************************************************************
*/



// IsControllableUnchecked(rPlantGen, rCAlph, rSupCandGen, rCriticalStates)
bool IsControllableUnchecked(
  const vGenerator& rPlantGen,
  const EventSet& rCAlph,  
  const vGenerator& rSupCandGen, 
  StateSet& rCriticalStates) 
{
  FD_DF("IsControllableUnchecked(" << &rSupCandGen << "," << &rPlantGen << ")");

  // HELPERS:

  // todo stack
  std::stack<Idx> todog, todoh;
  // set of already processed states
  StateSet processed;
  // reverse sorted transition relation of rSupCandGen (to be built on the fly)
  TransSetX2EvX1 rtransrel;

  // PREPARE RESULT:
  rCriticalStates.Clear();

  // ALGORITHM:

  // return false (uncontrollable) if there is no initial state
  if (rPlantGen.InitStatesEmpty() || rSupCandGen.InitStatesEmpty()) {
       return false;
  }

  // push combined initial state on todo stack
  todog.push(*rPlantGen.InitStatesBegin());
  todoh.push(*rSupCandGen.InitStatesBegin());
  FD_DF("IsControllable: todo push: (" << rPlantGen.SStr(*rPlantGen.InitStatesBegin()) << "|"
     << rSupCandGen.SStr(*rSupCandGen.InitStatesBegin()) << ")");

  // process todo stack
  while (! todog.empty()) {
    // get top element from todo stack
    Idx currentg = todog.top();
    Idx currenth = todoh.top();
    todog.pop();
    todoh.pop();
    FD_DF("IsControllable: todo pop: (" << rPlantGen.SStr(currentg) << "|" 
       << rSupCandGen.SStr(currenth) << ")");

#ifdef FAUDES_DEBUG_FUNCTION
    TransSet::Iterator _titg, _tith;
    // print all transitions of current states
    FD_DF("IsControllable: transitions from current states:");
    for (_titg = rPlantGen.TransRelBegin(currentg); _titg != rPlantGen.TransRelEnd(currentg); ++_titg) 
       FD_DF("IsControllable: g: " << rPlantGen.SStr(_titg->X1) << "-" 
		 << rPlantGen.EStr(_titg->Ev) << "-" << rPlantGen.SStr(_titg->X2));
    for (_tith = rSupCandGen.TransRelBegin(currenth); _tith != rSupCandGen.TransRelEnd(currenth); ++_tith) 
       FD_DF("IsControllable: h: " << rSupCandGen.SStr(_tith->X1) << "-" 
		 << rSupCandGen.EStr(_tith->Ev) << "-" << rSupCandGen.SStr(_tith->X2));
#endif 

    // process all h transitions while there could be matching g transitions
    TransSet::Iterator titg = rPlantGen.TransRelBegin(currentg);
    TransSet::Iterator titg_end = rPlantGen.TransRelEnd(currentg);
    TransSet::Iterator tith = rSupCandGen.TransRelBegin(currenth);
    TransSet::Iterator tith_end = rSupCandGen.TransRelEnd(currenth);
    while ((tith != tith_end) && (titg != titg_end)) {
      FD_DF("IsControllable: processing g-transition: " << rPlantGen.SStr(titg->X1) 
		  << "-" << rPlantGen.EStr(titg->Ev) << "-" << rPlantGen.SStr(titg->X2));
      FD_DF("IsControllable: processing h-transition: " << rSupCandGen.SStr(tith->X1) 
	          << "-" << rSupCandGen.EStr(tith->Ev) << "-" << rSupCandGen.SStr(tith->X2));
      // increment tith and titg case A: process common events
      if (titg->Ev == tith->Ev) {
	FD_DF("IsControllable: processing common event " << rPlantGen.EStr(titg->Ev));
	// add to todo list if state not processed // TODO: this may be incorrect !!!
	if (! processed.Exists(currenth)) {
	  todog.push(titg->X2);
	  todoh.push(tith->X2);
	  FD_DF("IsControllable: todo push: (" << rPlantGen.SStr(titg->X2) << "|"
	    << rSupCandGen.SStr(tith->X2) << ")");
	}
	// if h successor state is not critical add transition to rtransrel
	if (! rCriticalStates.Exists(tith->X2)) {
	  rtransrel.Insert(*tith);
	  FD_DF("IsControllable: incrementing g and h transrel");
	  ++titg;
	  ++tith;
	}
	// if successor state is critical and event uncontrollable 
	else if (!rCAlph.Exists(titg->Ev)) {
	  FD_DF("IsControllable: successor state " << rSupCandGen.SStr(tith->X2) << 
	    " critical and event " << rPlantGen.EStr(titg->Ev) << " uncontrollable:");
	  FD_DF("IsControllable: TraverseUncontrollableBackwards(" << rSupCandGen.SStr(currenth) << ")");
	    TraverseUncontrollableBackwards(rCAlph, rtransrel, rCriticalStates, currenth);
#ifdef FAUDES_CHECKED
	    // just increment transrel iterators to find all h transitions not in g
	    FD_DF("IsControllable: incrementing g an h transrel (FAUDES_CHECKED)");
	    ++titg;
	    ++tith;
	    break;
#else
            // exit all loops
	    titg = titg_end;
	    tith = tith_end;
	    break;
#endif
	}
        // if successor state is critical and event controllable
        else {
          FD_DF("IsControllable: incrementing g and h transrel");
          ++titg;
          ++tith;
        }		
      }
      // increment tith and titg case B: process g event that is not enabled in h
      else if (titg->Ev < tith->Ev) {
	FD_DF("IsControllable: asynchronous execution of event " 
	         << rPlantGen.EStr(titg->Ev) << " in g while " << rSupCandGen.EStr(tith->Ev)
		 << " in h");
        // if uncontrollable transition violates specification
        // delete state from rResGen and put into criticalset
        if (!rCAlph.Exists(titg->Ev)) {
	  FD_DF("IsControllable: asynchronous event " << rPlantGen.EStr(titg->Ev) 
		     << " in g is uncontrollable");
	  FD_DF("IsControllable: TraverseUncontrollableBackwards(" << rSupCandGen.SStr(currenth) << ")");
	  TraverseUncontrollableBackwards(rCAlph, rtransrel, rCriticalStates, currenth);
	  // exit all loops over g transrel
	  titg = titg_end;
	  break;
        }
	FD_DF("IsControllable: incrementing g transrel");
	++titg;
      }
      // increment tith and titg case C: process h event that is not enabled in g
      else {
       // #ifdef FAUDES_CHECKED
       // FD_WARN("IsControllable: transition " << rSupCandGen.TStr(*tith) 
       //  << " in specification h not found in g");
       // #endif
       FD_DF("IsControllable: incrementing h transrel");
       ++tith;
      }
    } // end while tith and titg
    // increment tith and titg case D: process leftover g events 
    while (titg != titg_end) {
      FD_DF("IsControllable: asynchronous execution of event " 
	        << rPlantGen.EStr(titg->Ev) << " in g at end of h");
      FD_DF("IsControllable: actual g-transition: " << rPlantGen.SStr(titg->X1) 
		<< "-" << rPlantGen.EStr(titg->Ev) << "-" << rPlantGen.SStr(titg->X2));
      FD_DF("IsControllable: actual h-transition: end");
      // if uncontrollable transition violates specification
      if (!rCAlph.Exists(titg->Ev)) {
	FD_DF("IsControllable: asynchronous execution of uncontrollable event "
		  << rPlantGen.EStr(titg->Ev) << " in g");
	FD_DF("IsControllable: TraverseUncontrollableBackwards(" << rPlantGen.SStr(currenth) << ")");
        TraverseUncontrollableBackwards(rCAlph, rtransrel, rCriticalStates, currenth);
	// exit this loop
	break;
      }
      FD_DF("IsControllable: incrementing g transrel");
      ++titg;
    }
    //#ifdef FAUDES_CHECKED
    // increment tith and titg case E: process leftover h events 
    //while (tith != tith_end) {
    //  FD_WARN("IsControllable: transition " << rSupCandGen.TStr(*tith) << " in specification h not found in g");
    //  FD_DF("IsControllable: incrementing h transrel");
    //  ++tith;
    //}
    //#endif
    processed.Insert(currenth);
  } //end while todog

  // compute overall set of critical states
  rCriticalStates = rSupCandGen.States() - ( processed - rCriticalStates );
	
  if (rCriticalStates.Empty()) {
    FD_DF("IsControllable: controllable");
    return true;
  }
  FD_DF("IsControllable: not controllable");
  return false;
}




// SupConUnchecked(rPlantGen, rCAlph, rSupCandGen)
bool SupConUnchecked(
  const vGenerator& rPlantGen, 
  const EventSet& rCAlph,
  vGenerator& rSupCandGen) 
{
  FD_DF("SupConUnchecked...");

  // HELPERS:

  // set of critical states
  StateSet critical;

  // ALGORITHM:
  
  // return true if plant or parallelcomp contain no initial states
  if(rPlantGen.InitStatesEmpty() || rSupCandGen.InitStatesEmpty()) {
    return true;
  }
   
  // test controllability
  IsControllable(rPlantGen, rCAlph, rSupCandGen, critical);
  FD_DF("SupCon: delete critical states: " << rSupCandGen.States().ToString());
  StateSet::Iterator sit;
  for(sit = critical.Begin(); sit != critical.End(); ++sit){
    rSupCandGen.DelState(*sit);
  }

  // return true if no states had to be deleted
  if(critical.Empty()) {
    FD_DF("SupCon: controllable");
    return true;
  }
  // return false if states had to be deleted
  FD_DF("SupCon: not controllable");
  return false;
}


// SupconParallel(rPlantGen, rCAlph, rSpecGen, rReverseCompositionMap, rResGen)
void SupconParallel(
  const vGenerator& rPlantGen, 
  const EventSet& rCAlph,
  const vGenerator& rSpecGen,
  std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
  vGenerator& rResGen) 
{
  FD_DF("SupconParallel(" << &rPlantGen << "," << &rSpecGen << ")");
	
  // HELPERS:

  // todo stack
  std::stack< std::pair<Idx,Idx> > todo;
  // set of critical states
  StateSet critical;
  // actual pair, new pair
  std::pair<Idx,Idx> currentstates, newstates;
  // other stuff
  Idx tmpstate;
  std::map< std::pair<Idx,Idx>, Idx>::iterator rcmapit;
  StateSet::Iterator lit1, lit2;
  TransSet::Iterator titg, titg_end, tith, tith_end;

  // ALGORITHM:
  
  if (rPlantGen.InitStatesEmpty()) {
    FD_DF("SupconParallel: plant got no initial states. "
            << "parallel composition contains empty language.");
    return;
  }
  if (rSpecGen.InitStatesEmpty()) {
    FD_DF("SupconParallel: spec got no initial states. "
              << "parallel composition contains empty language.");
    return;
  }

  // create initial state
  currentstates = std::make_pair(*rPlantGen.InitStatesBegin(), *rSpecGen.InitStatesBegin());
  todo.push(currentstates);
  rReverseCompositionMap[currentstates] = rResGen.InsInitState();
  FD_DF("SupconParallel: NEW ISTATE: (" << rPlantGen.SStr(currentstates.first) 
		  << "|" << rSpecGen.SStr(currentstates.second) << ") -> " 
                  << rReverseCompositionMap[currentstates]);
  if (rPlantGen.ExistsMarkedState(*rPlantGen.InitStatesBegin()) 
                && rSpecGen.ExistsMarkedState(*rSpecGen.InitStatesBegin())) {
    rResGen.SetMarkedState(rReverseCompositionMap[currentstates]);
     FD_DF("SupconParallel: recent ISTATE Marked");
  }

  // do parallel composition of allowed states by deleting critical states on the fly.
  // this creates an accessible generator
  FD_DF("SupconParallel: *** processing reachable states ***");
  while (! todo.empty()) {
    // get next reachable pair of states from todo stack
    currentstates = todo.top();
    todo.pop();
    FD_DF("SupconParallel: todo pop: (" << rPlantGen.SStr(currentstates.first) 
	       << "|" << rSpecGen.SStr(currentstates.second) << ") -> " 
               << rReverseCompositionMap[currentstates]);
		
    titg = rPlantGen.TransRelBegin(currentstates.first);
    titg_end = rPlantGen.TransRelEnd(currentstates.first);
    tith = rSpecGen.TransRelBegin(currentstates.second);
    tith_end = rSpecGen.TransRelEnd(currentstates.second);

#ifdef FAUDES_DEBUG_FUNCTION
    // print all transitions of current states
    FD_DF("SupConParallel: transitions from current states:");
    for (;titg != titg_end; ++titg) { 
      FD_DF("SupConParallel: g: " << rPlantGen.SStr(titg->X1) << "-" << rPlantGen.EStr(titg->Ev) 
	       << "-" << rPlantGen.SStr(titg->X2));
    }
    for (;tith != tith_end; ++tith) {
      FD_DF("SupConParallel: h: " << rSpecGen.SStr(tith->X1) << "-" << rSpecGen.EStr(tith->Ev) 
	        << "-" << rSpecGen.SStr(tith->X2));
    }
    titg = rPlantGen.TransRelBegin(currentstates.first);
    tith = rSpecGen.TransRelBegin(currentstates.second);
#endif                                                                      

    // process all h transitions while there could be matching g transitions
    while ((tith != tith_end) && (titg != titg_end)) {
      FD_DF("SupconParallel: current g-transition: " << rPlantGen.TStr(*titg) );
      FD_DF("SupconParallel: actual h-transition: " << rSpecGen.TStr(*tith));
      // increment titg and titg, case A: execute common events
      if (titg->Ev == tith->Ev) {
	FD_DF("SupconParallel: executing common event "  << rPlantGen.EStr(titg->Ev));
	newstates = std::make_pair(titg->X2, tith->X2);
	rcmapit = rReverseCompositionMap.find(newstates);
	// if state is new: add to todo list and result
	if (rcmapit == rReverseCompositionMap.end()) {
	  todo.push(newstates);
	  // if marked 
	  if (rPlantGen.ExistsMarkedState(newstates.first) && 
		rSpecGen.ExistsMarkedState(newstates.second)) {
	    tmpstate = rResGen.InsMarkedState();
	    FD_DF("SupconParallel: NEW MSTATE: (" << rPlantGen.SStr(newstates.first) << "|"
		     << rSpecGen.SStr(newstates.second) << ") -> " << tmpstate);
	  }
	  // if not maked 
	  else {
	    tmpstate = rResGen.InsState();
	    FD_DF("SupconParallel: NEW STATE: (" << rPlantGen.SStr(newstates.first) << "|" 
	             << rSpecGen.SStr(newstates.second) << ") -> " << tmpstate);
	  }
	  rReverseCompositionMap[newstates] = tmpstate;
	  FD_DF("SupconParallel: todo push: (" << rPlantGen.SStr(newstates.first)
		   << "|" << rSpecGen.SStr(newstates.second) << ") -> " << tmpstate);
	}
	// if state already exists
	else {
	  tmpstate = rcmapit->second;
	}
	// if successor state is not critical add transition and increment titg and tith
	if (! critical.Exists(tmpstate)) {
	  FD_DF("SupconParallel: ADDING TRANSITION " 
		   << rPlantGen.SStr(rReverseCompositionMap[currentstates]) << "-" << rPlantGen.EStr(titg->Ev) 
		   << "-" << rPlantGen.SStr(tmpstate));
	  rResGen.SetTransition(rReverseCompositionMap[currentstates], titg->Ev, tmpstate);
	  FD_DF("SupconParallel: incrementing g and h transrel");
	  ++titg;
	  ++tith;
	}
	// if successor state is critical and event is uncontrollable then delete current state 
	else if (!rCAlph.Exists(titg->Ev)) {
	  FD_DF("SupconParallel: successor " << rSpecGen.SStr(tmpstate) 
		       << "in critical and common event " << rSpecGen.EStr(titg->Ev) 
		       << " uncontrollable:");
	  FD_DF("SupconParallel: critical insert" << rPlantGen.SStr(tmpstate));
	  critical.Insert(tmpstate);
#ifdef FAUDES_CHECKED
	  // do not end while loops here for detecting all h transitions 
	  // not in g
	  FD_DF("SupconParallel: incrementing g and h transrel (FAUDES_CHECKED)");
	  ++titg;
	  ++tith;
#else
	  // exit all loops
	  titg = titg_end;
	  tith = tith_end;
#endif
        }
        // else if successor state in critical and event controllable increment titg and tith
        else {
	  FD_DF("SupconParallel: incrementing g and h transrel");
          ++titg;
          ++tith;
        }
      } // end: if processing common event 
      // increment titg and titg, case B: process g event that is not enebled for h
      else if (titg->Ev < tith->Ev) {
	FD_DF("SupconParallel: asynchronous execution of event " 
		 << rPlantGen.EStr(titg->Ev) << " in g while " << rSpecGen.EStr(tith->Ev) << " in h");
	// if uncontrollable transition leaves specification
	// delete state from res and put into criticalset
	if (!rCAlph.Exists(titg->Ev)) {
	  FD_DF("SupconParallel: asynchronous event " << rPlantGen.EStr(titg->Ev) 
	     << " in g is uncontrollable");
	  tmpstate = rReverseCompositionMap[currentstates];	
	  FD_DF("SupconParallel: critical insert" << rPlantGen.SStr(tmpstate));
	  critical.Insert(tmpstate);
	  // exit all loops
	  titg = titg_end;
	  tith = tith_end;
	  break;
	}
	FD_DF("SupconParallel: incrementing g transrel");
	++titg;
      } 
      // increment titg and titg, case C: process h event that is not enabled for g
      else {
	//#ifdef FAUDES_CHECKED
	// FD_WARN("SupconParallel:  transition " << rSpecGen.TStr(*tith) << " in specification h not found in g");
	//#endif
	FD_DF("SupconParallel: incrementing h transrel");
	++tith;
      }
    } // end while incrementing tith and titg 
    // increment titg and titg, case D: process leftover events of g
    while (titg != titg_end) {
      FD_DF("SupconParallel: asynchronous execution of event " 
	       << rPlantGen.EStr(titg->Ev) << " in g at end of h");
      FD_DF("SupconParallel: actual g-transition: " << rPlantGen.SStr(titg->X1) 
	       << "-" << rPlantGen.EStr(titg->Ev) << "-" << rPlantGen.SStr(titg->X2));
      FD_DF("SupconParallel: actual h-transition: end");
      // if uncontrollable transition leaves specification
      if (!rCAlph.Exists(titg->Ev)) {
	tmpstate = rReverseCompositionMap[currentstates];
	FD_DF("SupconParallel: asynchron executed uncontrollable end "
	         << "event " << rPlantGen.EStr(titg->Ev) << " leaves specification:");
	FD_DF("SupconParallel: critical insert" << rPlantGen.SStr(tmpstate));
	critical.Insert(tmpstate);
	// exit this loop 
	break;
      }
      FD_DF("SupconParallel: incrementing g transrel");
      ++titg;
    }
    //#ifdef FAUDES_CHECKED
    // increment titg and titg, case E: process leftover events of h
    //while (txoith != tith_end) {
    //  FD_WARN("SupconParallel:  transition " << rSpecGen.TStr(*tith)  << " in specification h not found in g");
    //  FD_DF("SupconParallel: incrementing h transrel");
    //  ++tith;
    //}
    //#endif
  } // while todo
  FD_DF("SupconParallel: deleting critical states...");
  StateSet::Iterator sit;
  for(sit = critical.Begin(); sit!= critical.End(); ++sit)
     rResGen.DelState(*sit); 
  //rResGen.DelStates(critical);
}



// SupConNBUnchecked(rPlantGen, rCAlph, rSpecGen, rReverseCompositionMap, rResGen)
void SupConNBUnchecked(
  const vGenerator& rPlantGen,
  const EventSet& rCAlph,  
  const vGenerator& rSpecGen,
  std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
  vGenerator& rResGen) 
{
  FD_DF("SupConNB(" << &rPlantGen << "," << &rSpecGen << ")");

  // PREPARE RESULT:	
  vGenerator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  pResGen->Clear();
  pResGen->Name(CollapsString("SupConNB(("+rPlantGen.Name()+"),("+rSpecGen.Name()+"))"));
  pResGen->InjectAlphabet(rPlantGen.Alphabet());

  // controllable events
  FD_DF("SupConNB: controllable events: "   << rCAlph.ToString());

  // ALGORITHM:
  SupconParallel(rPlantGen, rCAlph, rSpecGen, rReverseCompositionMap, *pResGen);

  // make resulting generator trim until it's fully controllable
  while (1) {
    if (pResGen->Empty()) {
      break;
    }
    Idx state_num = pResGen->Size();
    bool supcon = SupConUnchecked(rPlantGen, rCAlph, *pResGen);
    bool trim = pResGen->Trim();
    if (supcon && trim && (pResGen->Size() == state_num)) {
      break;
    }
  }

  // convenience state names
  if (rPlantGen.StateNamesEnabled() && rSpecGen.StateNamesEnabled() && rResGen.StateNamesEnabled()) 
    SetComposedStateNames(rPlantGen, rSpecGen, rReverseCompositionMap, *pResGen);
  else
    pResGen->ClearStateNames();

  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }

}

// ControlProblemConsistencyCheck(rPlantGen, rCAlph, rSpecGen)
void ControlProblemConsistencyCheck(
  const vGenerator& rPlantGen,
  const EventSet& rCAlph, 
  const vGenerator& rSpecGen) 
{

  // alphabets must match
  if ( rPlantGen.Alphabet() != rSpecGen.Alphabet()) {
    EventSet only_in_plant = rPlantGen.Alphabet() - rSpecGen.Alphabet();
    EventSet only_in_spec = rSpecGen.Alphabet() - rPlantGen.Alphabet();
    only_in_plant.Name("Only_In_Plant");
    only_in_spec.Name("Only_In_Specification");
    std::stringstream errstr;
    errstr << "Alphabets of generators do not match.";
    if(!only_in_plant.Empty())
      errstr << " " << only_in_plant.ToString() << ".";
    if(!only_in_spec.Empty())
      errstr << " " << only_in_spec.ToString() << ".";
    throw Exception("SupCon/SupConNB", errstr.str(), 100);
  }
	
  // TODO: add test for rCAlph <= Alphabet

  // TODO: check symboltables to match

  // plant and spec must be deterministic
  bool plant_det = rPlantGen.IsDeterministic();
  bool spec_det = rSpecGen.IsDeterministic();

  if ((plant_det == false) && (spec_det == true)) {
    std::stringstream errstr;
    errstr << "Plant generator must be deterministic, "
      << "but is nondeterministic";
    throw Exception("ControllableConsistencyCheck", errstr.str(), 201);
  }
  else if ((plant_det == true) && (spec_det == false)) {
    std::stringstream errstr;
    errstr << "Spec generator must be deterministic, "
       << "but is nondeterministic";
    throw Exception("ControllableConsistencyCheck", errstr.str(), 203);
  }
  else if ((plant_det == false) && (spec_det == false)) {
    std::stringstream errstr;
    errstr << "Plant and spec generator must be deterministic, "
       << "but both are nondeterministic";
    throw Exception("ControllableConsistencyCheck", errstr.str(), 204);
  }
}


/*
****************************************
* SUPCON: WRAPPER / USER FUNCTIONS     *
****************************************
*/


// IsControllable(rPlantGen, rCAlph, rSupCandGen)
bool IsControllable(
  const vGenerator& rPlantGen, 
  const EventSet& rCAlph, 
  const vGenerator& rSupCandGen) 
{
  FD_DF("IsControllable(" << &rSupCandGen << "," << &rPlantGen << ")");

  // HELPERS:
  FD_DF("IsControllable: controllable events: " << rCAlph.ToString());
  // critical set
  StateSet critical;

  // CONSISTENCY CHECKS:
  ControlProblemConsistencyCheck(rPlantGen, rCAlph, rSupCandGen);

  // ALGORITHM:
  return IsControllableUnchecked(rPlantGen, rCAlph, rSupCandGen, critical);
}


// IsControllable(rPlantGen, rCAlph, rSupCandGen, critical)
bool IsControllable(
  const vGenerator& rPlantGen, 
  const EventSet& rCAlph, 
  const vGenerator& rSupCandGen, 
  StateSet& rCriticalStates) 
{
  FD_DF("IsControllable(" << &rSupCandGen << "," << &rPlantGen << ")");

  // CONSISTENCY CHECKS:
  ControlProblemConsistencyCheck(rPlantGen, rCAlph, rSupCandGen);

  // ALGORITHM:
  return IsControllableUnchecked(rPlantGen, rCAlph, rSupCandGen, rCriticalStates);
}


// SupConNB(rPlantGen, rCAlph, rSpecGen, rResGen)
void SupConNB(
  const vGenerator& rPlantGen, 
  const EventSet& rCAlph, 
  const vGenerator& rSpecGen, 
  vGenerator& rResGen) 
{

  // CONSISTENCY CHECK:
  ControlProblemConsistencyCheck(rPlantGen, rCAlph, rSpecGen);

  // HELPERS:
  std::map< std::pair<Idx,Idx>, Idx> rcmap;

  // ALGORITHM:
  SupConNBUnchecked(rPlantGen, rCAlph, rSpecGen, rcmap, rResGen);
}




// SupConNBNonDet(rPlantGen, rCAlph, rSpecGen, rReverseCompositionMap, 
//     pDetPlantGen, pDetSpecGen, rResGen)
void SupConNBNonDet(
  const vGenerator& rPlantGen, 
  const EventSet& rCAlph,
  const vGenerator& rSpecGen,
  std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap,
  Generator rDetPlantGen, 
  Generator rDetSpecGen, 
  std::map<Idx,StateSet>& rPowerStatesPlant, 
  std::map<Idx,StateSet>& rPowerStatesSpec,
  bool& rDetPlantBool, 
  bool& rDetSpecBool, 
  vGenerator& rResGen) 
{
  FD_DF("SupConNBNonDet(" << &rPlantGen << "," << &rSpecGen << ")");

  // alphabets must match
  if (rPlantGen.Alphabet() != rSpecGen.Alphabet()) {
    throw Exception("SupConNBNonDet", "Alphabets of generators don't match", 100);
  }

  // PREPARE RESULT:
  vGenerator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  pResGen->Clear();
  pResGen->Name(CollapsString("SupConNBNonDet(("+rPlantGen.Name()+"),("+rSpecGen.Name()+"))"));
  pResGen->InjectAlphabet(rPlantGen.Alphabet());
	
  // HELPERS:
  rPowerStatesPlant.clear();
  rPowerStatesSpec.clear();

  // controllable events
  FD_DF("SupConNB: controllable events: " << rCAlph.ToString());

  // pointers to actual generators
  vGenerator* plantp;
  vGenerator* specp;

  // plant and spec determinism
  rDetPlantBool = rPlantGen.IsDeterministic();
  rDetSpecBool = rSpecGen.IsDeterministic();
	
  // create deterministic generators and set pointers to the actual ones
  plantp = const_cast<vGenerator*>(&rPlantGen);
  specp = const_cast<vGenerator*>(&rSpecGen);
  if (rDetPlantBool == false) {
      Deterministic(rPlantGen, rPowerStatesPlant, rDetPlantGen);		
      plantp = &rDetPlantGen;
  }
  if (rDetSpecBool == false) {
      Deterministic(rSpecGen, rPowerStatesSpec, rDetSpecGen);
      specp = &rDetSpecGen;
  }
  
  // ALGORITHM:

  // compute supervisor
  SupconParallel(*plantp, rCAlph, *specp, rReverseCompositionMap, *pResGen);

  // make resulting generator trim until it's fully controllable
  while (! (SupConUnchecked(*plantp, rCAlph, *pResGen) && pResGen->Trim()));

  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
}

// SupCon(rPlantGen, rCAlph, rSpecGen, rResGen)
void SupCon(
  const vGenerator& rPlantGen, 
  const EventSet& rCAlph,
  const vGenerator& rSpecGen, 
  vGenerator& rResGen) 
{
  // HELPERS:
  std::map< std::pair<Idx,Idx>, Idx> rcmap;

  // ALGORITHM:
  SupCon(rPlantGen, rCAlph, rSpecGen, rcmap, rResGen);
}


// SupCon(rPlantGen, rCAlph, rSpecGen, rReverseCompositionMap, rResGen)
void SupCon(
  const vGenerator& rPlantGen, 
  const EventSet& rCAlph,
  const vGenerator& rSpecGen,
  std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
  vGenerator& rResGen) 
{
  FD_DF("SupCon(" << &rPlantGen << "," << &rSpecGen << ")");

  // PREPARE RESULT:
	

  // prepare result
  vGenerator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }
  pResGen->Clear();
  pResGen->Name(CollapsString("SupCon(("+rPlantGen.Name()+"),("+rSpecGen.Name()+"))"));
  pResGen->InjectAlphabet(rPlantGen.Alphabet());
	
  // HELPERS:
  FD_DF("SupCon: controllable events: " << rCAlph.ToString());

  // CONSISTENCY CHECK:
  ControlProblemConsistencyCheck(rPlantGen, rCAlph, rSpecGen);

  // ALGORITHM:

  // parallel composition
  SupconParallel(rPlantGen, rCAlph, rSpecGen, rReverseCompositionMap, *pResGen);
  // make resulting generator controllable
  SupConUnchecked(rPlantGen, rCAlph, *pResGen);

  // convenience state names
  if (rPlantGen.StateNamesEnabled() && rSpecGen.StateNamesEnabled() && rResGen.StateNamesEnabled()) 
    SetComposedStateNames(rPlantGen, rSpecGen, rReverseCompositionMap, *pResGen);
  else
    pResGen->ClearStateNames();

  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
}


// TraverseUncontrollableBackwards(rCAlph, rtransrel, critical, current)
void TraverseUncontrollableBackwards(
  const EventSet& rCAlph, 
  TransSetX2EvX1& rtransrel, 
  StateSet& rCriticalStates, 
  Idx current) {
  FD_DF("TraverseUncontrollableBackwards: " << rCriticalStates.Str(current));

  // HELPERS:
  std::stack<Idx> todo;
  TransSetX2EvX1::Iterator rtit = rtransrel.BeginByX2(current);
  TransSetX2EvX1::Iterator rtit_end = rtransrel.EndByX2(current);

  // ALGORITHM:
  rCriticalStates.Insert(current);
  todo.push(current);
  FD_DF("TraverseUncontrollableBackwards: current rCriticalStates set: " 
     << rCriticalStates.ToString());
  // process todo stack
  while (! todo.empty()) {
    // get top state from todo stack
    current = todo.top(); todo.pop();
    // iteration over all transitions with x2 == current
    rtit_end = rtransrel.EndByX2(current);
    rtit = rtransrel.BeginByX2(current);
    for (; rtit != rtit_end; ++rtit) {
      // if uncontrollable event and predecessor state not already critical
      if ((!rCAlph.Exists(rtit->Ev)) && (! rCriticalStates.Exists(rtit->X1))) {
        FD_DF("TraverseUncontrollableBackwards: todo push " << rCriticalStates.Str(rtit->X1));
	todo.push(rtit->X1);
	FD_DF("TraverseUncontrollableBackwards: critical insert: " << rCriticalStates.Str(rtit->X1));
	rCriticalStates.Insert(rtit->X1);
      }
    }
  } // end while todo
}


// controllability test for cGenerator plant
bool IsControllable(
  const cGenerator& rPlantGen, 
  const vGenerator& rSupCandGen)
{
  return IsControllable(rPlantGen, rPlantGen.ControllableEvents(), rSupCandGen);
}


// supcon for cGenerators:
// uses and maintains controllablity from plant 
void SupConNB(
  const cGenerator& rPlantGen, 
  const vGenerator& rSpecGen, 
  vGenerator& rResGen) {

  // prepare result
  vGenerator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }

  // execute 
  SupConNB(rPlantGen, rPlantGen.ControllableEvents(),rSpecGen,*pResGen);

  // copy all attributes of input alphabet
  pResGen->EventAttributes(rPlantGen.Alphabet());

  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }

}

// supcon for cGenerators
// uses and maintains controllablity from plant 
void SupCon(
  const cGenerator& rPlantGen, 
  const vGenerator& rSpecGen, 
  vGenerator& rResGen) {

  // prepare result
  vGenerator* pResGen = &rResGen;
  if(&rResGen== &rPlantGen || &rResGen== &rSpecGen) {
    pResGen= rResGen.New();
  }

  // execute
  SupCon(rPlantGen, rPlantGen.ControllableEvents(),rSpecGen,*pResGen);

  // copy all attributes of input alphabet
  pResGen->EventAttributes(rPlantGen.Alphabet());

  // copy result
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }

}

} // name space 
