/** @file syn_supnorm.cpp Supremal normal sublanguage */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */
   

#include "syn_supnorm.h"


namespace faudes {




/*
***************************************************************************************
***************************************************************************************
 Implementation
***************************************************************************************
***************************************************************************************
*/


//void NormalityConsistencyCheck(rL,rOAlph,rK) 
void NormalityConsistencyCheck(
  const Generator& rL,
  const EventSet& rOAlph,
  const Generator& rK) {
  
  FD_DF("NormalityConsistencyCheck(...)");
  
  if (!(rK.IsDeterministic())) {
	 std::stringstream errstr;
     errstr << "Nondeterministic parameter rK.";
	 if(!(rL.IsDeterministic())) errstr << "Nondeterministic parameter rL.";
     throw Exception("NormalityConsistencyCheck", errstr.str(), 101);
	}
  if (!(rL.IsDeterministic())) {
	 std::stringstream errstr;
	 errstr << "Nondeterministic parameter rL.";
	 if(!(rL.IsDeterministic())) errstr << "Nondeterministic parameter rL.";
	 throw Exception("NormalityConsistencyCheck", errstr.str(), 101);
    }
  
  EventSet Kevents,Sigma;
  Sigma=rL.Alphabet();
  Kevents=rK.Alphabet();
  
  // observable events have to be subset of Sigma
  if(!(rOAlph<=Sigma)) {
    EventSet only_in_OAlph = rOAlph - Sigma;
    std::stringstream errstr;
    errstr << "Not all observable events are contained in Sigma: " 
       << only_in_OAlph.ToString() << ".";
    throw Exception("NormalityConsistencyCheck", errstr.str(), 100);
  } 
	 
  // alphabets must match  
  if ( Sigma != Kevents) {
    EventSet only_in_L = Sigma - Kevents;
    EventSet only_in_K = Kevents - Sigma;
    only_in_L.Name("Only_In_L");
    only_in_L.Name("Only_In_K");
    std::stringstream errstr;
    errstr << "Alphabets of generators do not match.";
    if(!only_in_L.Empty())
      errstr <<  " " << only_in_L.ToString() << ".";
    if(!only_in_K.Empty())
      errstr <<  " " << only_in_K.ToString() << ".";
    throw Exception("NormalityConsistencyCheck", errstr.str(), 100);
  }
  
  // K must be subset of L
  if (!LanguageInclusion(rK,rL)) {
    std::stringstream errstr;
    errstr << "K is not subset of L.";
    throw Exception("NormalityConsistencyCheck", errstr.str(), 0);
  }
}

// IsNormal(rK,rL,rOAlph)
bool IsNormal(
  const Generator& rL,
  const EventSet& rOAlph,
  const Generator& rK) 
{
  FD_DF("IsNormal(...)");
  
  NormalityConsistencyCheck(rL,rOAlph,rK);
  
	//extract overall alphabet:
  EventSet Sigma;
  Sigma=rL.Alphabet();
  
	//extract alphabet of rK:
  EventSet Kevents;
  Kevents=rK.Alphabet();
  
	//calculate p(K)
  Generator Ktest1, Ktest2;
  Project(rK,rOAlph,Ktest1);
  
	//calculate pinv(p(K))
  InvProject(Ktest1, Sigma);
  
	//check normality: pinv(p(K)) intersect L = K?
  LanguageIntersection(Ktest1,rL,Ktest2);
  return LanguageInclusion(Ktest2,rK);
	// Remark: LanguageEquality is not necessary, as inclusion
	// in the reverse direction is always met.
}

// IsNormalAlt(rK,rL,rOAlph)
bool IsNormalAlt(
  const Generator& rL,
  const EventSet& rOAlph,
    const Generator& rK) 
{
  FD_DF("IsNormalAlt(...)");
  
  NormalityConsistencyCheck(rL,rOAlph,rK);
  
	//extract overall alphabet:
  EventSet Sigma;
  Sigma=rL.Alphabet();
  
	//extract alphabet of rK:
  EventSet Kevents;
  Kevents=rK.Alphabet();
  
	//calculate p(K)
  Generator Ktest1, Ktest2;
  Project(rK,rOAlph,Ktest1);
  
	//calculate pinv(p(K))
  InvProject(Ktest1, Sigma);
  
  //check normality: p(K) || L = K?
  Parallel(Ktest1,rL,Ktest2);
  return LanguageInclusion(Ktest2,rK);
	// Remark: LanguageEquality is not necessary, as inclusion
	// in the reverse direction is always met.
}


// IsNormal rti wrapper
bool IsNormal(const cGenerator& rPlantGen, const vGenerator& rSupCandGen) {
  return IsNormal(rPlantGen, rPlantGen.ObservableEvents(),rSupCandGen);
}


// ConcatenateFullLanguage(rGen)
void ConcatenateFullLanguage(vGenerator& rGen) {
  FD_DF("ConcatenateFullLanguage(" << rGen.Name() << ")");
  
  // prepare result
  rGen.Name("ConcatenateFullLanguage(" + rGen.Name() + ")");
  
  // treat trivial empty result in case of empty marked language
  if(rGen.MarkedStatesSize()==0) {
		rGen.Clear();
		return;
	}
  
  // helpers
  EventSet alph=rGen.Alphabet();
  StateSet StatesToClear;
  TransSet TransToClear,TransToSet;
  StateSet::Iterator sit;
  TransSet::Iterator tit;
  EventSet::Iterator evit;
  
  // clear all transitions in marked states
  for (sit = rGen.MarkedStatesBegin(); sit != rGen.MarkedStatesEnd(); ++sit) {
      for (tit = rGen.TransRelBegin(*sit); tit != rGen.TransRelEnd(*sit); ++tit) {
          TransToClear.Insert(*tit);
      }
  }
  
  for (tit = TransToClear.Begin(); tit != TransToClear.End(); ++tit) {
      rGen.ClrTransition(*tit);
  }
  TransToClear.Clear();
  rGen.Accessible();
  
  // all marked states become eqivalent -> switch transitions leading to marked states to one remaining marked state
  // and delete all the others; note: for this special purpose, this procedure is faster than running StateMin.
  // if an init state is marked the effort is even less (as the result is Sigma* itself, see "else" below)
  bool initmarked = !( (rGen.InitStates() * rGen.MarkedStates()).Empty() );
  if (!initmarked) {
	  // define remaining marked state (there is one because of nonzero MarkedStatesSize)
	  sit = rGen.MarkedStatesBegin();
	  Idx marked = *sit;
	  // switch transitions to all other marked states if there are any
	  if(rGen.MarkedStatesSize()>1) {
		  // extract transitions sorted by target state X2
		  TransSetX2EvX1 trel;
		  rGen.TransRel(trel);
		  TransSetX2EvX1::Iterator tit2;
		  
		  // switch transitions to all other marked states to the one remaining marked state
		  ++sit;
		  for (; sit != rGen.MarkedStatesEnd(); ++sit) {
		      for (tit2 = trel.BeginByX2(*sit); tit2 != trel.EndByX2(*sit); ++tit2) {
				  // switch transition target to remaining marked state
				  TransToSet.Insert(tit2->X1,tit2->Ev,marked);
				  // note: original transitions are cleared later with clearing the corresp. marked states.
		      }
		  }
		  
		  // delete all but remaining marked state (note: by doing so, also corresp. transitions are cleared.)
		  sit = rGen.MarkedStatesBegin();
		  ++sit;
		  for (; sit != rGen.MarkedStatesEnd(); ++sit) {
			StatesToClear.Insert(*sit);
		  }
		  for (sit=StatesToClear.Begin(); sit != StatesToClear.End(); ++sit) {
			rGen.DelState(*sit);
		  }
		  StatesToClear.Clear();
		  
		  // insert switched transitions
		  for (tit = TransToSet.Begin(); tit != TransToSet.End(); ++tit) {
		      rGen.SetTransition(*tit);
		  }
		  TransToSet.Clear();
	    }

	  // now concatenate Sigma* by selflooping marked state with all events
	  for (evit = alph.Begin(); evit != alph.End(); ++evit) {
		  rGen.SetTransition(marked,*evit,marked);
	  }
	  
	}
  // consider marked init state case
  else {
		// reduce to empty string language
		rGen.ClearStates();
		rGen.ClearTransRel();
		Idx state=rGen.InsInitState();
		rGen.SetMarkedState(state);
		// now concatenate Sigma* by selflooping marked state with all events
		for (evit = alph.Begin(); evit != alph.End(); ++evit) {
		  rGen.SetTransition(state,*evit,state);
		}
		
    }
  
  return;
}


//  SupNorm(rL,rOAlph,rK,rResult)
bool SupNorm(
  Generator& rL,
  const EventSet& rOAlph,
  const Generator& rK, 
  Generator& rResult)
{
  FD_DF("SupNorm(" << rK.Name() << "," << rL.Name() << "," << rOAlph.Name() << ")");
  
  NormalityConsistencyCheck(rL,rOAlph,rK);
  
  // prepare result
  rResult.Clear();
  
  //extract overall alphabet:
  EventSet allevents;
  allevents=rL.Alphabet();
  
  //extract alphabet of rK:
  EventSet Kevents=rK.Alphabet();
  
  // remember name and marked states of rL
  std::string rLname=rL.Name();
  StateSet rLmarked=rL.MarkedStates();
  
  // involved operations from regular.h operate on the marked
  // languages -> turn generated languages into marked langs
  rL.InjectMarkedStates(rL.States());
  //###PrefixClosure(rL);
  Generator prK=rK;
  prK.InjectMarkedStates(prK.States());
  //###PrefixClosure(prK);
  rResult = prK;
  
  // calculate "L-K" (ToDo: use LanguageDifference):
  LanguageComplement(rResult);
  FD_DF("SupNorm: size of complement(K): "<<rResult.Size()<<",marked: "<<rResult.MarkedStatesSize());
  Generator tempgen;
  LanguageIntersection(rL, rResult, tempgen);
  rResult.Clear();
  FD_DF("SupNorm: sizeof L U complement(K)=L-K: "<<tempgen.Size());
  
  // calculate Pinv(P(L-K)):
  Project(tempgen,rOAlph,rResult);
  tempgen.Clear();
  FD_DF("SupNorm: sizeof p(L-K): "<< rResult.Size());
  InvProject(rResult,allevents);
  FD_DF("SupNorm: sizeof pinv(p(L-K)): "<<rResult.Size());
  
//  FD_DF("SupNorm: sizeof pinv(p(L-K)): "<<rResult.Size());

  //calculate remaining set difference -> supnorm(K)
  LanguageComplement(rResult);
  LanguageIntersection(prK, rResult, tempgen);
  rResult=tempgen;
  tempgen.Clear();
  FD_DF("SupNorm: sizeof K - pinv(p(L-K)) (result): "<<rResult.Size());
  
  // language MARKED by rResult is the result -> remove blocking states, mark all remaining states.
  rResult.Trim();
  rResult.InjectMarkedStates(rResult.States());
  
  // restore original rL
  rL.Name(rLname);
  rL.InjectMarkedStates(rLmarked);
  
  return !( rResult.InitStatesEmpty() );
}

//  SupNormClosed(rL,rOAlph,rK,rResult)
bool SupNormClosed(
  Generator& rL,
  const EventSet& rOAlph,
  const Generator& rK, 
  Generator& rResult)
{
  FD_DF("SupNormClosed(" << rK.Name() << "," << rL.Name() << "," << rOAlph.Name() << ")");
  
  NormalityConsistencyCheck(rL,rOAlph,rK);
  
  // prepare result
  rResult.Clear();
  
  //extract overall alphabet:
  EventSet allevents;
  allevents=rL.Alphabet();
  
  //extract alphabet of rK:
  EventSet Kevents=rK.Alphabet();
  
  // remember name and marked states of rL
  std::string rLname=rL.Name();
  StateSet rLmarked=rL.MarkedStates();
  
  // involved operations from regular.h operate on the marked
  // languages -> turn generated languages into marked langs
  rL.InjectMarkedStates(rL.States());
  //###PrefixClosure(rL);
  Generator prK=rK;
  prK.InjectMarkedStates(prK.States());
  //###PrefixClosure(prK);
  rResult = prK;
  
  // calculate "L-K" (ToDo: use LanguageDifference):
  LanguageComplement(rResult);
  FD_DF("SupNormClosed: size of complement(K): "<<rResult.Size()<<",marked: "<<rResult.MarkedStatesSize());
  Generator tempgen;
  LanguageIntersection(rL, rResult, tempgen);
  rResult.Clear();
  FD_DF("SupNormClosed: sizeof L U complement(K)=L-K: "<<tempgen.Size());
  
  // calculate Pinv(P(L-K)):
  Project(tempgen,rOAlph,rResult);
  tempgen.Clear();
  FD_DF("SupNormClosed: sizeof p(L-K): "<< rResult.Size());
  InvProject(rResult,allevents);
  FD_DF("SupNormClosed: sizeof pinv(p(L-K)): "<<rResult.Size());
  
  //concatenate Pinv(P(L-K)) with Sigma*: this leads to closed result
  ConcatenateFullLanguage(rResult); // special function for fast concatenation of Sigma*.
  
  //  FD_DF("SupNormClosed: sizeof pinv(p(L-K))Sigma*: "<<rResult.Size());

  //calculate remaining set difference -> supnormClosed(K)
  LanguageComplement(rResult);
  LanguageIntersection(prK, rResult, tempgen);
  rResult=tempgen;
  tempgen.Clear();
  FD_DF("SupNormClosed: sizeof K - pinv(p(L-K))Sigma* (result): "<<rResult.Size());
  
  // language MARKED by rResult is the result -> remove blocking states, mark all remaining states.
  rResult.Trim();
  rResult.InjectMarkedStates(rResult.States());
  
  // restore original rL
  rL.Name(rLname);
  rL.InjectMarkedStates(rLmarked);
  
  // mark result corresponding to rL
  LanguageIntersection(rResult,rL,rResult);
  
  return !( rResult.InitStatesEmpty() );
}



// SupPrefixClosed(rK,rResult)
bool SupPrefixClosed(
  const Generator& rK,
	Generator& rResult) {
  
	FD_DF("SupPrefixClosed("<<rK.Name()<<")");
  
	// prepare Result:
  rResult.Clear();
  rResult.Name("SupPrefixClosed("+rK.Name()+")");
  
  // check for marked initial state, empty result if not
  if( (rK.InitStates() * rK.MarkedStates()).Empty() ) return false;
  
  rResult=rK;  
  
  // erase all transitions not leading to a marked state
  // todo: Depth-first-search could be faster
  TransSet trel=rResult.TransRel();
  TransSet::Iterator tit=trel.Begin();
  TransSet::Iterator tit_end=trel.End();
  for(;tit!=tit_end;tit++) {
    if(!(rResult.ExistsMarkedState(tit->X2))) {
		  rResult.ClrTransition(*tit);
			}
  }
  
  // make reachable (cosmetic)
  rResult.Trim();
  
  // as there is at least one marked init state, result is nonempty
  return true;
}

// SupNormAlt(rK,rL,rOAlph,ClosedResult,rResult)
bool SupNormAlt(
  Generator& rL,
  const EventSet& rOAlph,
  const Generator& rK, 
  Generator& rResult) 
{
 	FD_DF("SupNorm(...)");
  
  NormalityConsistencyCheck(rL,rOAlph,rK);
  
  // prepare result
  rResult.Clear();
  rResult = rK;
  
  //extract overall alphabet:
  EventSet allevents;
  allevents=rL.Alphabet();
  
  //extract alphabet of rK:
  EventSet Kevents;
  Kevents=rK.Alphabet();
  
  // remember name and marked states of rL
  std::string rLname=rL.Name();
  StateSet rLmarked=rL.MarkedStates();
  
  // involved operations from regular.h treat only the marked
  // languages -> turn generated languages into marked langs
  rResult.InjectMarkedStates(rResult.States());
  //###PrefixClosure(rResult);
  rL.InjectMarkedStates(rL.States());
  //##PrefixClosure(rL);
  Generator Kquer=rK;
  Kquer.InjectMarkedStates(Kquer.States());
  //###PrefixClosure(Kquer);
  
  // calculate "L-K" (ToDo: use LanguageDifference):
  LanguageComplement(rResult);
  FD_DF("SupNorm: size of complement(K): "<<rResult.Size()<<",marked: "<<rResult.MarkedStatesSize());
  Generator tempgen;
  LanguageIntersection(rL, rResult, tempgen);
  rResult.Clear();
  FD_DF("SupNorm: sizeof L U complement(K)=L-K: "<<tempgen.Size());
  
  // calculate Pinv(P(L-K)):
  Project(tempgen,rOAlph,rResult);
  tempgen.Clear();
  FD_DF("SupNorm: sizeof p(L-K): "<< rResult.Size());
  InvProject(rResult,allevents);
  FD_DF("SupNorm: sizeof pinv(p(L-K)): "<<rResult.Size());
  
  // Remark: different from the above supnorm version, we do not
  // concatenate Sigma* here, but make the result prefix-closed
  // at the end.

  //calculate remaining set difference
  LanguageComplement(rResult);
  LanguageIntersection(Kquer, rResult, tempgen);
  rResult=tempgen;
  FD_DF("SupNorm: sizeof K - pinv(p(L-K))Sigma* (result): "<<rResult.Size());
  
  // make prefix-closed
  SupPrefixClosed(tempgen,rResult);

  // language MARKED by rResult is the result -> turn marked language into generated language
  rResult.Trim(); //may be unnecessesary
  
  // restore original rL
  rL.Name(rLname);
  rL.InjectMarkedStates(rLmarked);
  
  // mark result corresponding to rL
  LanguageIntersection(rResult,rL,rResult);
  
  return rResult.Empty();

}



/** rti wrapper */
void SupNorm(
  const vGenerator& rPlantGen, 
  const EventSet&  rOAlph,
  const vGenerator& rSpecGen, 
  vGenerator& rResGen) {

  // trust deferred copy ...
  vGenerator pg=rPlantGen;
  SupNorm(pg,rOAlph,rSpecGen,rResGen);

}

/** rti wrapper */
void SupNorm(
  const cGenerator& rPlantGen, 
  const vGenerator& rSpecGen, 
  vGenerator& rResGen) {

  // trust deferred copy ...
  vGenerator pg=rPlantGen;
  SupNorm(pg,rPlantGen.ObservableEvents(),rSpecGen,rResGen);

}

/** rti wrapper */
void SupNormClosed(
  const vGenerator& rPlantGen, 
  const EventSet&  rOAlph,
  const vGenerator& rSpecGen, 
  vGenerator& rResGen) {

  // trust deferred copy ...
  vGenerator pg=rPlantGen;
  SupNormClosed(pg,rOAlph,rSpecGen,rResGen);

}

/** rti wrapper */
void SupNormClosed(
  const cGenerator& rPlantGen, 
  const vGenerator& rSpecGen, 
  vGenerator& rResGen) {

  // trust deferred copy ...
  vGenerator pg=rPlantGen;
  SupNormClosed(pg,rPlantGen.ObservableEvents(),rSpecGen,rResGen);

}

} // end namespace
