/** @file cfl_helper.cpp   Helper functions */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Copyright (C) 2008-2010 Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA 

*/


#include "cfl_helper.h"


// Debug includes
#include "cfl_attributes.h"
#include "cfl_registry.h"
#include "cfl_types.h"
#include "cfl_elementary.h"

/* Include extra defs for mswindows */
#ifdef FAUDES_WINEXTRA
#include "cfl_winextra.h"
#endif


namespace faudes {

// ToStringInteger(number)
std::string ToStringInteger(long int number) { 
  if(number>= std::numeric_limits<long int>::max()) return "inf";
  if(number<= std::numeric_limits<long int>::min()+1) return "-inf";
  std::string res;
  std::stringstream sstr;
  sstr << number;
  sstr >> res;
  return res;
}

// ToStringInteger16(number)
std::string ToStringInteger16(long int number) { 
   std::string res;
   std::stringstream sstr;
   sstr << "0x" << std::setbase(16) << number;
   sstr >> res;
   return res;
}

// ToStringFloat(number)
// todo: check range, prevent sci notation
std::string ToStringFloat(double number) { 
   if(number == (long int) number) 
     return(ToStringInteger((long int) number));
   std::string res;
   std::stringstream sstr;
   sstr << std::fixed;
   sstr << number;
   sstr >> res;
   return res;
}

// ExpandString(rString, len) 
std::string ExpandString(const std::string& rString, unsigned int len) {
  std::string res;
  res	= rString;
  std::string::size_type xtra = (std::string::size_type) len - rString.length();
  if ((xtra > 0) && (xtra < 10000)) {
	  res.append(xtra, ' ');
	}
  return res;
} 

// CollapseString(rString, len) 
std::string  CollapsString(const std::string& rString, unsigned int len) {
  if(len <=1) return rString;
  if(rString.length() <= len) return rString;
  int chead = len/2;
  int ctail = len-chead;
  return rString.substr(0,chead) + "..." + rString.substr(rString.length()-ctail,ctail);
} 

// ToIdx(rString)
Idx ToIdx(const std::string& rString) {
	char * end;
	unsigned long ul = strtoul (rString.c_str(), &end, 0);
	unsigned long idxmax = std::numeric_limits<Idx>::max();
	if (ul > idxmax) {
		throw Exception("atoidx", "Idx overflow", 600);
	}
	else {
		return (Idx) ul;
	}
}

// String Substitution
std::string StringSubstitute(const std::string& rString,const std::string& rFrom,const std::string& rTo) {
  // prep result
  std::string res;
  std::size_t pos=0; 
  // loop over occurences of "from"
  while(pos<rString.length()) {
    std::size_t next=rString.find(rFrom,pos);
    if(next==std::string::npos) break;
    res.append(rString.substr(pos, next));
    res.append(rTo);
    pos=next+rFrom.length();
  }
  // get end
  if(pos<rString.length()) 
    res.append(rString.substr(pos));
  // done
  return res;
}

// FDVersionString()
std::string FDVersionString() {
	return std::string(FAUDES_VERSION);
}

// FDPluginsString()
std::string FDPluginsString() {
  return std::string(FAUDES_PLUGINS);
}

// FDContributorsString()
std::string FDContributorsString() {
  return 
    "Ruediger Berndt, Christian Breindl, Tobias Barthel, Christoph Doerr, Marc Duevel, Norman Franchi, Rainer Hartmann, Jochen Hellenschmidt, Andreas Mohr, Thomas Moor, Mihai Musunoi, Bernd Opitz, Irmgard Petzoldt, Sebastian Perk, Thomas Rempel, Daniel Ritter, Berno Schlein, Klaus Schmidt, Matthias Singer, Thomas Wittmann, Jorgos Zaddach, et al";
}


// ProcessDot(infile,outfile,format)
void ProcessDot(const std::string& rDotFile, 
  const std::string& rOutFile, const std::string& rOutFormat, const std::string& rDotExec)
{
  std::string format=rOutFormat;
  // guess format from filename
  if(format=="") {
    if(rOutFile.rfind('.')+1 < rOutFile.size()) {
       format=rOutFile.substr(rOutFile.rfind('.')+1);
    }
  }  
  // test format
  if (format == "canon"); 
  else if (format == "dot"); 
  else if (format == "xdot"); 
  else if (format == "cmap");
  else if (format == "dia");
  else if (format == "fig"); 
  else if (format == "gd"); 
  else if (format == "gd2"); 
  else if (format == "gif"); 
  else if (format == "hpgl"); 
  else if (format == "imap"); 
  else if (format == "cmapx");
  else if (format == "ismap"); 
  else if (format == "jpg"); 
  else if (format == "jpeg"); 
  else if (format == "mif"); 
  else if (format == "mp"); 
  else if (format == "pcl"); 
  else if (format == "pic"); 
  else if (format == "plain"); 
  else if (format == "plain-ext"); 
  else if (format == "png"); 
  else if (format == "ps"); 
  else if (format == "ps2"); 
  else if (format == "svg"); 
  else if (format == "svgz"); 
  else if (format == "vrml"); 
  else if (format == "vtx"); 
  else if (format == "wbmp"); 
  else {
    std::stringstream errstr;
    errstr << "Dot output format \"" << format << "\" unknown";
    throw Exception("faudes::ProcessDot", errstr.str(), 3);
  }
  std::string dotcommand = rDotExec + " -T"+format+" \""+rDotFile+"\" -o \""+rOutFile+"\"";
  if(system(dotcommand.c_str()) != 0) {
    throw Exception("faudes::ProcessDot", 
        "Error in running " + dotcommand, 3);
  }
}


// CreateTempFile(void)
// todo: sys dependant, report, investigate threads
std::string CreateTempFile(void) {
 char filename[]= "faudes_temp_XXXXXX";
  int filedes = mkstemp(filename);
  if(filedes==-1) {
    FD_DF("faudes::CreateTempFile(): error");
    return "";
  }
  close(filedes);
  std::string res(filename);

  FD_DF("faudes::CreateTempFile(): " << res);
  return(res);
}


// RemoveFile(void)
// todo: sys dependant *
bool RemoveFile(const std::string& rFileName) {
  return std::remove(rFileName.c_str()) == 0;
}


#ifndef FAUDES_DIR_SEPARATORS 
#define FAUDES_DIR_SEPARATORS "/"
#endif

// ExtractPath
std::string ExtractDirectory(const std::string& rFullPath) {
  std::string res="";
  unsigned int seppos = rFullPath.find_last_of(FAUDES_DIR_SEPARATORS);
  if(seppos==std::string::npos) return res;
  res=rFullPath.substr(0,seppos+1);
  return res;
}

// ExtractFilename
std::string ExtractFilename(const std::string& rFullPath) {
  std::string res=rFullPath;
  unsigned int seppos = rFullPath.find_last_of(FAUDES_DIR_SEPARATORS);
  if(seppos==std::string::npos) return res;
  res=rFullPath.substr(seppos+1);
  return res;
}

// PrependPath
std::string PrependDirectory(const std::string& rDirectory, const std::string& rFileName) {
  std::string res=rDirectory;
  char sepchar=std::string(FAUDES_DIR_SEPARATORS).at(0);
  if(res!="")
  if(res.at(res.length()-1)!=sepchar)
    res.append(1,sepchar);
  res.append(rFileName);
  return res;
}

// ConsoleOut class
ConsoleOut::ConsoleOut(void) : pStream(NULL),mMute(false) {
}
ConsoleOut::~ConsoleOut(void) {
  if(pStream) { pStream->flush(); delete pStream; }
  if(this==spInstance) spInstance=NULL;
}
ConsoleOut* ConsoleOut::G(void) {
  if(!spInstance) spInstance= new ConsoleOut();
  return spInstance;
}
void ConsoleOut::Redirect(ConsoleOut* out) {
  std::string fname;
  if(spInstance) { 
    fname = spInstance->Filename();
    spInstance->ToFile("");
  }
  spInstance=out; 
  if(!spInstance) spInstance=this;
  spInstance->ToFile(fname);
}
void ConsoleOut::ToFile(const std::string& filename) {
  if(pStream) { pStream->flush(); delete pStream; }
  pStream=NULL;
  mFilename=filename;
  if(mFilename=="") return;
  pStream = new std::ofstream();
  pStream->open(mFilename.c_str(),std::ios::app);
}
void ConsoleOut::Write(const std::string& message) {
  if(mMute) return;
  Buffer() << message;
  Flush();
}
std::ostream& ConsoleOut::Buffer(void) {
  return mLineBuffer;
}
void ConsoleOut::Flush(void) {
  if(!mMute) DoWrite(mLineBuffer.str());
  mLineBuffer.str("");
}
void ConsoleOut::DoWrite(const std::string& message) {
  std::ostream* sout=pStream;
  if(!sout) sout=&std::cerr;
  *sout << message;
  sout->flush();
}

// global instance
ConsoleOut* ConsoleOut::spInstance=NULL;
 


// debugging: objectcount
std::map<std::string,long int>* ObjectCount::mspMax=NULL;
std::map<std::string,long int>* ObjectCount::mspCount=NULL;
bool ObjectCount::msDone=false;
ObjectCount::ObjectCount(void) {
  mspCount= new std::map<std::string,long int>();
  mspMax= new std::map<std::string,long int>();
  msDone=true; 
}
void  ObjectCount::Init(void) {
 if(!msDone) ObjectCount();
}
void  ObjectCount::Inc(const std::string& rTypeName) {
  if(!msDone) ObjectCount();
  long int cnt = ((*mspCount)[rTypeName]+=1);
  if((*mspMax)[rTypeName]<cnt) (*mspMax)[rTypeName]=cnt;
}
void  ObjectCount::Dec(const std::string& rTypeName) {
  if(!msDone) ObjectCount();
  (*mspCount)[rTypeName]-=1;
}


// debugging: report on exit function
void ExitFunction(void){
#ifdef FAUDES_DEBUG_CODE
  FAUDES_WRITE_CONSOLE("faudes::ExitFunction:");
  // be sure its up and running
  ObjectCount::Init();
  // get rid of all registry prototypes
  TypeRegistry::G()->Clear();  
  // prepare report
  std::map<std::string,long int>::iterator cit;
  cit=ObjectCount::mspCount->begin();
  for(;cit!=ObjectCount::mspCount->end();cit++) {
    FAUDES_WRITE_CONSOLE( cit->first << ": #" << ToStringInteger(cit->second) <<
     " (max #" << (*ObjectCount::mspMax)[cit->first] << ")");
  }
#endif
}
 

#ifdef FAUDES_DEBUG_CODE
// report on exit install
class ExitFunctionInstall {
private:
  static bool mDone;
  static ExitFunctionInstall mInstance;
  ExitFunctionInstall(void) {
    if(mDone) return;
    FAUDES_WRITE_CONSOLE("ExitFunctionInstall()");
    std::atexit(ExitFunction);
    mDone=true;
  } 
};
// exit function: global data
bool ExitFunctionInstall::mDone=false;
ExitFunctionInstall ExitFunctionInstall::mInstance;
#endif

// test protocol: token writer
TokenWriter* gTestProtocolTw=NULL;

// test protocol: setup
void TestProtocol(const std::string& rSource) {
  if(gTestProtocolTw) delete gTestProtocolTw;
  std::string filename=rSource;
  // fix empty source
  if(filename=="") filename="test_dump";
  // remove directory
  filename=ExtractFilename(filename);
  // remove extension
  std::string::size_type pos=0;
  for(;pos<filename.length();pos++) 
    if(filename.at(pos)=='.') filename.at(pos)='_';
  // append extension
  filename.append(".prot");
  // prepend prefix
  filename.insert(0,"tmp_");
  // initialise token writer
  gTestProtocolTw= new TokenWriter(filename);
}
  
// test protocol: dump
void TestProtocol(const std::string& rSource, const std::string& rMessage, const Type& rData, bool core) {
  if(!gTestProtocolTw) TestProtocol(rSource);
  gTestProtocolTw->WriteComment("%%% test mark: " + rMessage);
  if(core) rData.Write(*gTestProtocolTw);
  else rData.SWrite(*gTestProtocolTw);
  gTestProtocolTw->WriteComment("");
  gTestProtocolTw->WriteComment("");
  gTestProtocolTw->WriteComment("");
  *gTestProtocolTw << "\n";
}
void TestProtocol(const std::string& rSource, const std::string& rMessage, bool data) {
  Boolean fbool(data);
  TestProtocol(rSource,rMessage,fbool,true);
} 
void TestProtocol(const std::string& rSource, const std::string& rMessage, long int data) {
  Integer fint(data);
  TestProtocol(rSource,rMessage,fint,true);
} 
void TestProtocol(const std::string& rSource, const std::string& rMessage, const std::string& rData) {
  String fstr(rData);
 TestProtocol(rSource,rMessage,fstr,true);
} 


// declare loop callback
static bool (*gBreakFnct)(void)=0;

// set loop callback
void LoopCallback( bool pBreak(void)) {
  gBreakFnct=pBreak;
}

// do loop callback
void LoopCallback(void){
  if(!gBreakFnct) return;
  if(! (*gBreakFnct)() ) return;
  throw Exception("LoopCallback", "break on application request", 110);
}

} // namespace faudes
