/** @file cfl_token.cpp @brief Class Token */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2006  Bernd Opitz
Exclusive copyright is granted to Klaus Schmidt

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "cfl_token.h"

namespace faudes {


// Token::Token()
Token::Token(void) : 
  mType(None), 
  mStringValue(""), 
  mIntegerValue(0), 
  mFloatValue(0)
{ 
}

// copy construct
Token::Token(const Token& rToken) :
  mType(rToken.mType), 
  mStringValue(rToken.mStringValue), 
  mIntegerValue(rToken.mIntegerValue), 
  mFloatValue(rToken.mFloatValue)
{ 
}


// assignment
Token& Token::operator=(const Token& rToken) {
  mType=rToken.mType; 
  mStringValue=rToken.mStringValue; 
  mIntegerValue=rToken.mIntegerValue; 
  mFloatValue=rToken.mFloatValue;
  return *this;
}

// destructor
Token::~Token(void){
}

// Token::SetNone()
void Token::SetNone(void ) {
  mType=None; mStringValue=""; mIntegerValue=0; mFloatValue=0;
}

// Token::SetString(str)
void Token::SetString(const std::string& rName) {
  mType=String; mStringValue=rName; mIntegerValue=0; mFloatValue=0;
}

// Token::SetBegin(str)
void Token::SetBegin(const std::string& rName) {
  mType=Begin; mStringValue=rName; mIntegerValue=0; mFloatValue=0;
}

// Token::SetEnd(str)
void Token::SetEnd(const std::string& rName) {
  mType=End; mStringValue=rName; mIntegerValue=0; mFloatValue=0;
}

// Token::SetOption(str)
void Token::SetOption(const std::string& rName) {
  mType=Option; mStringValue=rName; mIntegerValue=0; mFloatValue=0;
}

// Token::SetInteger(number)
void Token::SetInteger(const long int number) {
  mType=Integer; mStringValue=""; mIntegerValue=number; mFloatValue=number;
}

// Token::SetInteger16(number)
void Token::SetInteger16(const long int number) {
  mType=Integer16; mStringValue=""; mIntegerValue=number; mFloatValue=number;
}

// Token::SetFloat(number)
void Token::SetFloat(const double number) {
  mType=Float; mStringValue=""; mIntegerValue=(long int) number; mFloatValue=number;
}

// Token::SetBinary
void Token::SetBinary(const char* data, long int len) {
  mStringValue.assign(data,len);
  mType=Binary;
}


// ReadNumber(pStream)
bool Token::ReadNumber(std::istream* pStream){
  char c, vc;
  double fv=0;
  long int iv=0;
  int comma = -1;
  bool minus = false;
  int  base=10;
  bool ok=false;
  int cnt=0;
  // check the whole pStream
  while(*pStream){
    // check eof
    if (pStream->eof()) 
      {mType=None; return(false);}; 
    // check next charakter
    c = pStream->peek();                 
    cnt++;
    // break if space
    if (isspace(c)) break;
    // change base on x
    if(c=='x' && iv==0 && cnt==2 && comma<0 && !minus) 
      {pStream->get(); base = 16; continue;}
    // change sign on -
    if(c=='-' && cnt==1) 
      {pStream->get(); minus = true; continue;}
    // record comma
    if(c=='.' && comma<0 && base==10) 
      {pStream->get(); comma = 0; continue;}
    // read +infinity
    if(c=='i' && cnt==1) {
      ReadString(pStream,' ');
      if(mStringValue=="inf") 
        {iv= std::numeric_limits<long int>::max(); fv= iv; break;}
    }
    // read -infinity
    if(c=='i' && cnt==2 && minus) {
      ReadString(pStream,' ');
      if(mStringValue=="inf") 
        {iv=std::numeric_limits<long int>::min()+1; fv=iv; break;}
    }
    // break if it is not a digit
    if (!isdigit(c) && base==10) break;
    if (!isxdigit(c) && base==16) break;
    // get the next charakter
    pStream->get();
    // compute the value of c
    vc=0;
    if(c>='0' && c<= '9') vc = c-'0';
    else if(c>='a' && c<= 'f') vc = c-'a' + 10;
    else if(c>='A' && c<= 'F') vc = c-'A' + 10;
    // compute the corresponding number
    iv = base * iv + vc;
    fv = base * fv + vc;
    if(comma>=0) comma++;
    ok = true; 
  } // termination with the end of the pStream or a non-digit or a space
  // fix sign
  if(minus) {
    iv=-iv;
    fv=-fv;
  }
  // fix decimal point
  for(;comma>0;comma--) {
    iv/=base;
    fv/=base;
  }
  // assign the numeric value and type in Token
  mIntegerValue = iv;
  mFloatValue = fv;
  mType= (comma>0 || minus) ? Float : Integer;
  if(base == 16) mType = Integer16;
  return(ok);
}

// Write(pStream)
void Token::Write(std::ostream* pStream){
  FD_DV("Token::Write: mType=" << (int) mType 
	<< "\" mStringValue=\"" << mStringValue 
	<< "\" mIntegerValue=" <<mIntegerValue 
	<< "\" mFloatValue=" <<mFloatValue <<"\n");
  // distuingish the token type
  switch (mType) {
  // <mStringValue>	
  case Begin:
    *pStream << '<' << mStringValue << '>';
    break;
  // </mStringValue>
  case End:
    *pStream << "</" << mStringValue << '>';
    break;
  // "mStringValue"
  case String:
     WriteString(pStream,'"');
     break;
  // +mStringValue+
  case Option:
    *pStream <<  ExpandString("+"+mStringValue+"+", FD_NAMELEN);
    break;
  // mIntegerValue
  case Integer:
    *pStream << ExpandString(ToStringInteger(mIntegerValue), FD_NAMELEN);
    break;
  // mIntegerValue
  case Integer16:
    *pStream << ExpandString(ToStringInteger16(mIntegerValue), FD_NAMELEN);
    break;
  // mFloatValue
  case Float:
    *pStream << ExpandString(ToStringFloat(mFloatValue), FD_NAMELEN);
    break;
  // =BinaryData=
  case Binary:
     WriteBinary(pStream);
     break;
  // all other unexpected cases
  default:
    assert(0);
  }
}

// WriteString(pStream, char)
void Token::WriteString(std::ostream* pStream, char stop){
  // assemble escape character string
  std::string escstr="\\";
  escstr.append(1,stop);
  std::string stopstr;
  stopstr.append(1,stop);
  // no escape characters
  if(mStringValue.find_first_of(escstr)==std::string::npos) {
    *pStream <<  ExpandString(stopstr+mStringValue+stopstr, FD_NAMELEN);
    return;
  }
  // do escape substitution
  *pStream << stopstr;
  std::string::iterator cit=mStringValue.begin(); 
  while(cit!=mStringValue.end()) {
    if(escstr.find_first_of(*cit)!=std::string::npos) *pStream << '\\';
    *pStream << *cit;
    cit++;
  }
  *pStream << stopstr;
  *pStream << " ";
}

// ReadString(pStream, char)
int Token::ReadString(std::istream* pStream, char stop){
  int lc=0;
  char c;
  bool escape=false;
  bool ctrlblank = false;
  mStringValue = "";
  // check the whole pStream
  while (*pStream) {
    // check eof
    if (pStream->eof()) {mType=None; return -1;}; 
    // get one character
    c = pStream->get();
    // count the next lines
    if (c == '\n') ++lc;
    //replace sequence of control characters by one blanc
    if(iscntrl(c) && ctrlblank) continue;
    ctrlblank=false;
    if(iscntrl(c)) { c=' '; ctrlblank=true;}
    // if not in escape mode ... 
    if(!escape) {
      // ... sense stop
      if(c == stop) break;
      // ... sense escape
      if(c == '\\') escape=true;
      // ... add character
      if(c != '\\') mStringValue.append(1,c);
      continue;
    }
    // else we are in escape mode: add verb and clear
    mStringValue.append(1,c);
    escape=false;
  } 
  return lc;
}


// WriteVerbatim(pStream)
  void Token::WriteVerbatim(std::ostream* pStream, const std::string& rString){
  // verb marker candidate
  std::string verb="__VERBATIM__";
  // adjust marker
  while(rString.find(verb)!=std::string::npos) {
    verb = verb+"#";
  }
  // do write
  *pStream << std::endl;
  *pStream << verb << std::endl;
  *pStream << rString;
  *pStream << verb << std::endl;
}


// ReadString(pStream, char)
int Token::ReadVerbatim(std::istream* pStream){
  // clear
  mStringValue = "";
  std::string verb="_";
  int lc=0;
  char c;
  // figure mark
  while(*pStream) {
    // check eof
    if(pStream->eof()) {mType=None; return -1;}; 
    // get one character
    c = pStream->get();
    // new line, control and space  breaks
    if(c == '\n')  { lc++; break; }
    if(iscntrl(c)) break;
    if(isblank(c)) break;
    // others get appended
    verb.append(1,c);
  }
  // read until mark
  while (*pStream) {
    // check eof
    if (pStream->eof()) {mType=None; return -1;}; 
    // get one character
    c = pStream->get();
    // count the next lines
    if (c == '\n') ++lc; 
    // append anything
    mStringValue.append(1,c);
    // if c is _ or #, test for end
    if(c== verb.at(verb.size()-1)) { 
      unsigned int vpos=mStringValue.find(verb);
      if(vpos!=std::string::npos) {
        mStringValue = mStringValue.substr(0,vpos);
        return lc;
       }
    }
  }
  return lc;
}

// WriteBinary(pStream,data,len)
void Token::WriteBinary(std::ostream* pStream,  const char* pData, long int len) {

  // my encoding (hardcoded in read, however)
  static char Base64EncodingTable[]=
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

  // start
  *pStream << "=";

  // loop vars
  const char* src = pData;
  long int cnt=len;
  int step=0;
  unsigned char c0=0,c1=0,c2=0,c3=0;

  // encode and output buffer
  while(cnt>0) {
    switch(step) {
    // collect char for 1st byte
    case 0:
      c0=  ((*src & 0xFC) >> 2);
      c1=  ((*src & 0x03) << 4);
      c2=0;
      c3=0;
      step=1;
      break;
    // collect char for 2nd byte
    case 1:
      c1|= ((*src & 0xF0) >> 4);
      c2|= ((*src & 0x0F) << 2);
      step=2;
      break;
    // collect char for 3rd byte, plus output
    case 2:
      c2|= ((*src & 0xC0) >> 6);
      c3|= (*src & 0x3F); 
      *pStream << Base64EncodingTable[c0] << Base64EncodingTable[c1] << 
        Base64EncodingTable[c2]  << Base64EncodingTable[c3]; 
      step=0;
      break;
    default: break;
    }
    cnt--;
    src++;
  }   
  // flush left overs, incl padding
  switch(step) {
   case 0:
     *pStream << "= ";
     break;
   case 1:
     *pStream << Base64EncodingTable[c0] << Base64EncodingTable[c1] << "=== ";
     break;
   case 2:
     *pStream << Base64EncodingTable[c0] << Base64EncodingTable[c1] << 
       Base64EncodingTable[c2]  << "== ";
     break;
  }
}

// WriteBinary(pStream)
void Token::WriteBinary(std::ostream* pStream){
  if(mType!=Binary) return;
  WriteBinary(pStream,mStringValue.data(),mStringValue.size());
}


// ReadBinary ... to mStringValue
int Token::ReadBinary(std::istream* pStream) {
  
  // line count
  int lc;

  // read as string, excl. marks '='
  lc=ReadString(pStream,'=');

  // take extra trailing padding =
  while(!pStream->eof()) {
    if(pStream->peek()!='=') break;
    pStream->get();
    mStringValue.append("=");
  }

  // loop vars
  std::string::iterator src=mStringValue.begin();
  std::string::iterator dst=mStringValue.begin();
  long int cnt=0;
  int c=0, d0=0, d1=0, d2=0;
  unsigned char cs[4];
  unsigned int step=0;

  // loop
  while(true) {
    
    // get data
    while(true) {
      // sense eof
      if(src==mStringValue.end()) { c = -1; break; }
      // get char
      c=*(src++);
      // decode
      if(c>='A' && c <='Z') { c-='A'; break; }
      if(c>='a' && c <='z') { c-='a'; c+= ('Z'-'A')+1; break; }
      if(c>='0' && c <='9') { c-='0'; c+= 2*('Z'-'A')+2; break; }
      if(c=='+') {c= 62; break; }
      if(c=='/') {c= 63; break; }
      if(c=='=') {c= 0xFF; break; };
    }
    // pass on eof
    if(c== -1) break;
    // record and continue
    cs[step++] = c;
    if(step<=3) continue;
    step=0;
    // sort bits
    d0= ((cs[0] << 2) & 0xFC) | ((cs[1] >> 4) & 0x03);
    d1= ((cs[1] << 4) & 0xF0) | ((cs[2] >> 2) & 0x0F);
    d2= ((cs[2] << 6) & 0xC0) | (cs[3] & 0x3F);
    // record result
    if(cs[0]!= 0xFF && cs[1]!=0xFF) {*(dst++)=d0; cnt++;}
    if(cs[1]!= 0xFF && cs[2]!=0xFF) {*(dst++)=d1; cnt++;}
    if(cs[2]!= 0xFF && cs[3]!=0xFF) {*(dst++)=d2; cnt++;}
    // sense end
    if(cs[3]==0xFF) break;
  }   

  // set data length (sets the length incl termination char??)
  mStringValue.resize(cnt);
 
  // return padding error or line count
  return step==0 ? lc : -1;
}



// ReadSpace(pStream)
int Token::ReadSpace(std::istream* pStream){
  char c = '\0';
  int lc = 0;
  // check the whole pStream
  while (*pStream) {
    while (*pStream) {
      // check eof
      if (pStream->eof()) {return(lc);}; 
      // look one character ahead
      c = pStream->peek();
      // break if the next character is 
      if((! isspace(c)) && (! iscntrl(c))) break;
      // not a space or a control character
      // get the next space or control character
      pStream->get();
      // count the next lines
      if(c == '\n') ++lc;
    }
    // if the next charakter starts a comment
    if (c != '%') break;
    while (*pStream) {
      // check eof
      if (pStream->eof()) {return(lc);}; 
      // get the next character
      c = pStream->get();
      // count the next lines
      if (c == '\n') ++lc;
      // terminate with the next new line  character
      if (c == '\n') break;
      if (c == '\r') break;
    }
  }
  // termination if the next character is not a space, a control character 
  //or a '%'.
  return (lc); 
}

// Read(pStream)
int Token::Read(std::istream* pStream){
  char c1, c2;
  int lc = 0;
  int ll;
  // the token is initialized with the type "None"
  SetNone();
  // check eof
  if (pStream->eof()) return(lc); 
  // read all spaces in the pStream
  lc += ReadSpace(pStream);
  // check eof
  if (pStream->eof()) return(lc);
  // get the first useful character
  c1 = pStream->peek();
  if (isdigit(c1) || (c1=='-') || (c1=='i') ) {
    // token is a number if the character is a digit, a sign or i(nfinity)
    ReadNumber(pStream);
  } 
  // token is a string if it starts with '"'
  else if (c1 == '"') {
    pStream->get();
    // read the string until '"'
    ll=ReadString(pStream,'"');
    lc+=ll;
    if(ll>=0) mType = String;
  } 
  // token is an option string if it starts with '+'
  else if (c1 == '+') {
    pStream->get();
    // read the string until '+'
    ll=ReadString(pStream,'+');
    lc+=ll;
    if(ll>=0) mType = Option;
  } 
  // token is a binary string if it starts with '='
  else if (c1 == '=') {
    pStream->get();
    // read the string until '=', incl padding
    ll=ReadBinary(pStream);
    lc+=ll;
    if(ll>=0) mType = Binary;
  } 
  // token is a string if it starts with '_'
  else if (c1 == '_') {
    pStream->get();
    // read the string until the corresponding end marker
    ll=ReadVerbatim(pStream);
    lc+=ll;
    if(ll>=0) mType = String;
  } 
  // token is of type begin or end if it starts with <
  else if (c1 == '<') {
    pStream->get();
    // check eof
    if (pStream->eof()) return(lc); 
    c2 = pStream->peek(); 
    // token is of type "begin" if it starts with <
    if(c2 != '/') {
      if(ReadString(pStream,'>')>=0) mType = Begin;
    } 
    else {  
      // token is of type "end" if it starts with </
      pStream->get();
      // read the string until '>'
      if(ReadString(pStream,'>')>=0) mType = End;
    }
  } 
  FD_DV("Token::Read: char1='"  << c1 << "' mType=" << (int) mType 
	<< "\" mStringValue=\"" << mStringValue 
	<< "\" mIntegerValue="  << mIntegerValue 
	<< "\" mFloatValue="    << mFloatValue <<"\n");
  return(lc);
}

} // namespace faudes
