/** @file cfl_types.cpp Runtime interface, faudes types */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2009 Ruediger Berndt
Copyright (C) 2010 Thomas Moor

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "cfl_types.h"

namespace faudes{

/*
********************************************************************
********************************************************************
********************************************************************

implemantation of faudes Type

********************************************************************
********************************************************************
********************************************************************
*/

// static members: ref to the only one instance
const std::string Type::mName("TypeStdName");

// constructor
Type::Type(void)  {}

// copy constructor
Type::Type(const Type& rType)  {(void) rType;}

// destructor
Type::~Type(void) {}

// pointer constructor
Type* Type::New(void) const { 
  return new Type(); 
}

// pointer copy constructor
Type* Type::Copy(void) const { 
  return new Type(*this); 
}

// cast
const Type* Type::Cast(const Type* pOther) const {
  return pOther;
}

// Clear()
void Type::Clear(void) {
  FD_DC("Type::Clear(): not re-implemented in " << typeid(*this).name());
}

// assign
Type& Type::Assign(const Type& rSource) {
  FD_DC("Type(" << this << ")::Assign(" << &rSource << ")");
  Clear();
  return *this;
}

// equality
bool Type::Equal(const Type& rOther) const { 
  return true;
}   

// equality
bool Type::operator==(const Type& rOther) const { 
  return true;
}

// equality
bool Type::operator!=(const Type& rOther) const { 
  return false;
}

// assign
Type& Type::operator=(const Type& rSource) {
  FD_DC("Type(" << this << ")::AssignementOperator(" << &rSource << ")");
  Clear();
  return *this;
}

// assign
Type& Type::DoAssign(const Type& rSource) {
  FD_DC("Type(" << this << ")::DoAssign(" << &rSource << ")");
  return *this;
}

// equality
bool Type::DoEqual(const Type& rOther) const {
  FD_DC("Type(" << this << ")::DoEqual(" << &rOther << ")");
  return true;
}

// Name(rName)
void Type::Name(const std::string& rName) {
}

// Name()
const std::string& Type::Name(void) const {
  return mName;
}

// Write(context)
void Type::Write(const Type* pContext) const {
  TokenWriter tw(TokenWriter::Stdout);
  DoWrite(tw,"",pContext);
}

// Write(rFileName, label, context, openmode)
void Type::Write(const std::string& rFileName, const std::string& rLabel, 
		 const Type* pContext, std::ios::openmode openmode) const {
  try {
    TokenWriter tw(rFileName, openmode);
    DoWrite(tw, rLabel, pContext);
  }
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/writing file \"" << rFileName << "\"";
    throw Exception("Type::Write", errstr.str(), 2);
  }
}

// Write(rFileName, openmode)
void Type::Write(const std::string& rFileName, std::ios::openmode openmode) const {
  Write(rFileName,"",0,openmode);
}

// Write(tw,label,context)
void Type::Write(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  DoWrite(rTw,rLabel,pContext);
}

// ToString(label, context)
std::string Type::ToString(const std::string& rLabel, const Type* pContext) const {
  TokenWriter tw(TokenWriter::String);
  tw.Endl(false);
  DoWrite(tw,rLabel,pContext);
  return tw.Str();
}

// ToSText()
std::string Type::ToSText(void) const {
  TokenWriter tw(TokenWriter::String);
  tw.Endl(true);
  DoSWrite(tw);
  return tw.Str();
}

// ToString(label, context)
std::string Type::ToText(const std::string& rLabel, const Type* pContext) const {
  TokenWriter tw(TokenWriter::String);
  tw.Endl(true);
  DoWrite(tw,rLabel,pContext);
  return tw.Str();
}

// DWrite(context)
void Type::DWrite(const Type* pContext) const {
  TokenWriter tw(TokenWriter::Stdout);
  DoDWrite(tw,"",pContext);
}

// DWrite(rFileName, label, context, openmode)
void Type::DWrite(const std::string& rFileName, const std::string& rLabel, 
		  const Type* pContext, std::ios::openmode openmode) const {
  try {
    TokenWriter tw(rFileName, openmode);
    DoDWrite(tw, rLabel, pContext);
  }
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/writing file \"" << rFileName << "\"";
    throw Exception("Type::DWrite", errstr.str(), 2);
  }
}

// DWrite(tw,label,context)
void Type::DWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  DoDWrite(rTw,rLabel,pContext);
}

// SWrite()
void Type::SWrite(void) const {
  TokenWriter tw(TokenWriter::Stdout);
  DoSWrite(tw);
}

// SWrite(tw)
void Type::SWrite(TokenWriter& rTw) const {
  DoSWrite(rTw);
}

// Read(rFilename, rLabel,context)
void Type::Read(const std::string& rFilename, const std::string& rLabel, const Type* pContext) {
  Clear();
  TokenReader tr(rFilename);
  DoRead(tr,rLabel,pContext);
}

// Read(rTr, rLabel, context)
void Type::Read(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  Clear();
  DoRead(rTr,rLabel,pContext);
}

// FromString(rString, rLabel, context)
void Type::FromString(const std::string& rString, const std::string& rLabel, const Type* pContext) {
  Clear();
  TokenReader tr(TokenReader::String,rString);
  DoRead(tr,rLabel,pContext);
}
  
//DoWrite(rTr,rLabel,pContext)
void Type::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) rTw; (void) rLabel; (void) pContext;
  FD_DC("Type::DoWrite(): not re-implemented in " << typeid(*this).name());
}

//DoDWrite(rTr,rLabel,pContext)
void Type::DoDWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  FD_DC("Type::DoDWrite(): not re-implemented in " << typeid(*this).name() << ", using DoDWrite instead");
  DoWrite(rTw,rLabel,pContext);
}

//DoSWrite(rTr,rLabel,pContext)
void Type::DoSWrite(TokenWriter& rTw) const {
  FD_DC("Type::DoSWrite(): not re-implemented in " << typeid(*this).name());
  rTw.WriteComment("");
  rTw.WriteComment(" Statistics for " + Name());
  rTw.WriteComment("");
}

//DoRead(rTr,rLabel,pContext)
void Type::DoRead(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  (void) rLabel; (void) pContext; (void) rTr;
  FD_DC("Type::DoRead(): not re-implemented in " << typeid(*this).name());
}



/*
********************************************************************
********************************************************************
********************************************************************

Implementation of class Documentation

********************************************************************
********************************************************************
********************************************************************
*/

// faudes type
FAUDES_TYPE_IMPLEMENTATION(Documentation,Type)

// construct
Documentation::Documentation(void) : Type() {
}

// construct
Documentation::Documentation(const Documentation& rOther) : Type() {
  DoAssign(rOther);
}

// std faudes type
Documentation& Documentation::DoAssign(const Documentation& rSrc) {
  // assign my members
  mName=rSrc.mName;
  mPlugIn=rSrc.mPlugIn;
  mHtmlDoc=rSrc.mHtmlDoc;
  mTextDoc=rSrc.mTextDoc;
  mKeywords=rSrc.mKeywords;
  return *this;
}

// std faudes type
bool Documentation::DoEqual(const Documentation& rOther) const {
  // test my members
  if(mName!=rOther.mName) return false;
  if(mPlugIn!=rOther.mPlugIn) return false;
  if(mHtmlDoc!=rOther.mHtmlDoc) return false;
  if(mTextDoc!=rOther.mTextDoc) return false;
  if(mKeywords!=rOther.mKeywords) return false;
  return true;
}

// clear all
void Documentation::Clear(void){
  FD_DRTI("Documentation::Clear()");
  mName="";
  mPlugIn="CoreFaudes";
  mHtmlDoc="";
  mTextDoc="";
  mKeywords="";
}

//  read access members
const std::string& Documentation::Name(void) const{ return mName; }
const std::string& Documentation::PlugIn(void) const{ return mPlugIn; }
const std::string& Documentation::CType(void) const{ return mCType; }
const std::string& Documentation::TextDoc() const{ return mTextDoc; }
const std::string& Documentation::HtmlDoc() const{ return mHtmlDoc; }
const std::string& Documentation::Keywords() const{ return mKeywords; }

//  write access members
void Documentation::Name(const std::string& name){mName = name; }
void Documentation::PlugIn(const std::string& plugin){mPlugIn = plugin; }
void Documentation::CType(const std::string& name){mCType = name; }
void Documentation::TextDoc(const std::string& textdoc){mTextDoc = textdoc;}
void Documentation::HtmlDoc(const std::string& fname){mHtmlDoc = fname;}

//  write access keyword
void Documentation::AddKeyword(const std::string& rKeyword){
  if(mKeywords.empty()){
    mKeywords = rKeyword;
  } else{
    mKeywords.append(1, mDelim);
    mKeywords.append(rKeyword);
  }
}

// regex match keywords (todo)
std::string Documentation::MatchKeyword(const std::string& rPattern) const{
  FD_DRTI("Documentation::MatchKeyword(" << rPattern << ")");
  // be case insensitive (inefficient)
  std::string keys = mKeywords;
  std::transform(keys.begin(), keys.end(), keys.begin(), tolower);
  std::string match = rPattern;
  std::transform(match.begin(), match.end(), match.begin(), tolower);
  // find match
  std::string res="";
  std::size_t posk=keys.find(match);
  if(posk==std::string::npos) return res;
  // find delimiters 
  std::size_t posa=0;
  std::size_t posb=keys.length();
  while(true) {
    posb=keys.find(mDelim,posa);
    if(posb==std::string::npos) posb=mKeywords.length();
    if(posb>posk) break;
    posa=posb+1;
  };
  // return original case match
  FD_DRTI("Documentation::MatchKeyword(" << rPattern << "): " << mKeywords.substr(posa,posb-posa));
  return mKeywords.substr(posa,posb-posa);
}


// positional keyword
std::string Documentation::KeywordAt(int pos) const{
  FD_DRTI("Documentation::KeywordAt(" << pos << ")");
  // default result
  std::string res="";
  // find delimiter
  std::size_t posa=0;
  std::size_t posb=mKeywords.length();
  while(pos>0) {
    if(posa>=mKeywords.length()) return res;
    posb=mKeywords.find(mDelim,posa);
    if(posb==std::string::npos) return res;
    posa=posb+1;
    pos--;
  };
  // default result
  res=mKeywords.substr(posa);
  // find next delimiter
  posb=res.find(mDelim);
  if(posb==std::string::npos) return res;
  res=res.substr(0,posb);
  return res;
}

// keyword
int Documentation::KeywordsSize(void) const{
  int cnt=0;
  // find delimiter
  std::size_t pos=0;
  while(pos<mKeywords.length()) {
    cnt++;
    pos=mKeywords.find(mDelim,pos);
    if(pos==std::string::npos) break;
    pos++;
  };
  return cnt;
}


// merge with docu from token io
void Documentation::MergeDocumentationBody(TokenReader& rTr) {
  FD_DRTI("Documentation::MergeDocumentationBody(): file " << rTr.FileName() << " for " << mName);
  // peek type ident
  Token token;
  rTr.Peek(token);
  // separate name
  std::string name = token.StringValue();
  size_t pos=name.find("::");
  if(pos!=std::string::npos) name=name.substr(pos+2);
  // match?
  if(mName!=name && mName!="") {
    std::stringstream errstr;
    errstr << "Documentation mismatch in file \"" << rTr.FileName() << "\" : " << mName << "!=" << name;
    throw Exception("Documentation::MergeDocumentationBody", errstr.str(), 48);
  };
  // read core
  DoReadCore(rTr);  
}

// token io
void Documentation::DoRead(TokenReader& rTr,  const std::string& rLabel, const Type* pContext) {
  FD_DRTI("Documentation::DoRead(): file " << rTr.FileName());
  // ignore
  (void) pContext;
  // figure ma section
  std::string label = rLabel;
  if(label=="") label="Documentation";
  // doit
  rTr.ReadBegin(label);
  // peek and set name
  Token token;
  rTr.Peek(token);
  // extract name 
  mName=token.StringValue();
  size_t pos=mName.find("::");
  if(pos!=std::string::npos) mName=mName.substr(pos+2);
  // do read
  DoReadCore(rTr);
  // done
  rTr.ReadEnd(label);
}

// token read
void Documentation::DoReadCore(TokenReader& rTr) {
  FD_DRTI("Documentation::DoReadCore()");
  Token token;
  rTr.Peek(token);
  // my faudes entity (eg type name or function name)
  mName=token.StringValue();
  size_t pos=mName.find("::");
  if(pos!=std::string::npos) mName=mName.substr(pos+2);
  // my faudes plugin
  mPlugIn=token.StringValue();
  pos=mPlugIn.find("::");
  if(pos!=std::string::npos) mPlugIn=mPlugIn.substr(0,pos);
  else mPlugIn = "CoreFaudes";
  // flush and token check
  rTr.ReadString();
  // corresponding c type
  mCType="";
  rTr.Peek(token);
  if(token.Type()==Token::Option)
    mCType = rTr.ReadOption();
  // loop other known entries
  while(true) {
    rTr.Peek(token);
    // do the read: text
    if(token.Type()==Token::Begin)  
    if(token.StringValue()=="TextDoc") {
      rTr.ReadBegin("TextDoc");
      mTextDoc = rTr.ReadString();
      rTr.ReadEnd("TextDoc");
      continue;
    }
    // do the read: html
    if(token.Type()==Token::Begin)  
    if(token.StringValue()=="HtmlDoc") {
      rTr.ReadBegin("HtmlDoc");
      mHtmlDoc = rTr.ReadString();
      rTr.ReadEnd("HtmlDoc");
      continue;
    }
    // do the read: keys
    if(token.Type()==Token::Begin)  
    if(token.StringValue()=="Keywords") {
      rTr.ReadBegin("Keywords");
      while(!rTr.Eos("Keywords")) 
        AddKeyword(rTr.ReadString());
      rTr.ReadEnd("Keywords");
      continue;
    }
    // unknown -> derived class take over
    break;
  }
}


// token write
void Documentation::DoWrite(TokenWriter& rTw,  const std::string& rLabel, const Type* pContext) const {
  // ignore
  (void) pContext;
  // figure my section
  std::string label = rLabel;
  if(label=="") label="Documentation";
  // doit
  rTw.WriteBegin(label);
  DoWriteCore(rTw); 
  rTw.WriteEnd(label);
}

// token write
void Documentation::DoWriteCore(TokenWriter& rTw) const {
  // write name
  rTw << mPlugIn + "::" + mName ;
  // write c type
  if(mCType!="") rTw.WriteOption(mCType);
  // write text doc
  rTw.WriteBegin("TextDoc");
  rTw << mTextDoc;
  rTw.WriteEnd("TextDoc");
  // write html doc
  rTw.WriteBegin("HtmlDoc");
  rTw << mHtmlDoc;
  rTw.WriteEnd("HtmlDoc");
  // write keys
  rTw.WriteBegin("Keywords");
  rTw << mKeywords;
  rTw.WriteEnd("Keywords");
}



/*
********************************************************************
********************************************************************
********************************************************************

Implementation of class TypeDefinition

********************************************************************
********************************************************************
********************************************************************
*/


// typedefinition constructor function
TypeDefinition* TypeDefinition::Constructor(Type* pProto, const std::string& rTypeName){
  FD_DRTI("TypeDefinition::Construct(" << typeid(pProto).name() << ", " << rTypeName << ")");
  TypeDefinition* td = new TypeDefinition();
  td->Prototype(pProto);
  std::string name=rTypeName;
  if(name=="") name=typeid(pProto).name();
  td->Name(name);
  return(td);
}


// token io
void TypeDefinition::DoRead(TokenReader& rTr,  const std::string& rLabel, const Type* pContext) {
  FD_DRTI("TypeDefinition::DoRead(): file " << rTr.FileName());
  // ignore
  (void) pContext;
  // figure my section
  std::string label = rLabel;
  if(label=="") label="TypeDefinition";
  // base can handle this
  Documentation::DoRead(rTr,label,pContext);
}

// token io
void TypeDefinition::DoReadCore(TokenReader& rTr) {
  FD_DRTI("TypeDefinition::DoReadCore(): file " << rTr.FileName());
  // call base core
  Documentation::DoReadCore(rTr);  
  // no extra data for type definition
}

// token io
void TypeDefinition::DoWrite(TokenWriter& rTw,  const std::string& rLabel, const Type* pContext) const {
  // label
  std::string label=rLabel;
  if(label=="") label="TypeDefinition";
  // base can handle
  Documentation::DoWrite(rTw,label,pContext); 
}

// token io
void TypeDefinition::DoWriteCore(TokenWriter& rTw) const {
  FD_DRTI("TypeDefinition::DoWriteCore(): file " << rTw.FileName());
  // call base core
  Documentation::DoWriteCore(rTw);  
  // no extra data for type definition
}

// access prototype
const Type* TypeDefinition::Prototype(void) const{
  return(mpType);
}

// set prototype
void TypeDefinition::Prototype(faudes::Type* pType){
  if(mpType) delete mpType;
  mpType = pType;
}

// construct new object with faudes type
Type* TypeDefinition::NewObject() const{
  FD_DRTI("TypeDefinition::NewObject()");
  // bail out 
  if(!mpType) return NULL;
  // use prototype
  return mpType->New();
}


// clear (all except prototype)
void TypeDefinition::Clear(){
  FD_DRTI("TypeDefinition::Clear()");
  // call base
  Documentation::Clear();
}




} //namspace

